#!/usr/bin/python
# coding: UTF-8
#
# Author:   Dawid Laszuk
# Contact:  https://github.com/laszukdawid/PyEMD/issues
#
# Edited:   11/05/2017
#
# Feel free to contact for any information.

import logging
import time

import numpy as np
from scipy.interpolate import interp1d

from .splines import akima
from .utils import unify_types


class EMD:
    """
    Empirical Mode Decomposition

    *Note:*
    Default and recommended package for EMD is EMD.py.
    This is meant to provide with the same results as MATLAB version of EMD,
    which is not necessarily the most efficient or numerically accurate.

    Method of decomposing signal into Intrinsic Mode Functions (IMFs)
    based on algorithm presented in Huang et al. [1].

    Algorithm was validated with Rilling et al. [2] Matlab's version from 3.2007.

    [1] N. E. Huang et al., "The empirical mode decomposition and the
        Hilbert spectrum for non-linear and non stationary time series
        analysis", Proc. Royal Soc. London A, Vol. 454, pp. 903-995, 1998
    [2] G. Rilling, P. Flandrin and P. Goncalves, "On Empirical Mode
        Decomposition and its algorithms", IEEE-EURASIP Workshop on
        Nonlinear Signal and Image Processing NSIP-03, Grado (I), June 2003
    """

    logger = logging.getLogger(__name__)

    def __init__(self):
        self.splineKind = "cubic"

        self.nbsym = 2
        self.reduceScale = 1.0
        self.maxIteration = 500
        self.scaleFactor = 100

        self.FIXE = 0
        self.FIXE_H = 0

        self.stop1 = 0.05
        self.stop2 = 0.5
        self.stop3 = 0.05

        self.DTYPE = np.float64
        self.MAX_ITERATION = 1000

        self.TIME = False

    def extractMaxMinSpline(self, T, S):
        """
        Input:
        -----------------
            S - Input signal array. Should be 1D.
            T - Time array. If none passed numpy arange is created.

        Output:
        -----------------
            maxSpline - Upper envelope of signal S.
            minSpline - Bottom envelope of signal S.
            maxExtrema - Position (1st row) and values (2nd row) of maxima.
            minExtrema - Position (1st row) and values (2nd row) of minima.
        """

        # Get indexes of extrema
        maxPos, maxVal, minPos, minVal, _ = self.findExtrema(T, S)

        if len(maxPos) + len(minPos) < 3:
            return [-1] * 4

        # Extrapolation of signal (ober boundaries)
        maxExtrema, minExtrema = self.preparePoints(S, T, maxPos, maxVal, minPos, minVal)

        _, maxSpline = self.splinePoints(T, maxExtrema, self.splineKind)
        _, minSpline = self.splinePoints(T, minExtrema, self.splineKind)

        return maxSpline, minSpline, maxExtrema, minExtrema

    def preparePoints(self, S, T, maxPos, maxVal, minPos, minVal):
        """
        Adds to signal extrema according to mirror technique.
        Number of added points depends on nbsym variable.

        Input:
        ---------
            S: Signal (1D numpy array).
            T: Timeline (1D numpy array).
            maxPos: sorted time positions of maxima.
            maxVal: signal values at maxPos positions.
            minPos: sorted time positions of minima.
            minVal: signal values at minPos positions.

        Output:
        ---------
            minExtrema: Position (1st row) and values (2nd row) of minima.
            minExtrema: Position (1st row) and values (2nd row) of maxima.
        """

        # Find indices for time array of extrema
        indmin = np.array([np.nonzero(T == t)[0] for t in minPos]).flatten()
        indmax = np.array([np.nonzero(T == t)[0] for t in maxPos]).flatten()

        # Local variables
        nbsym = self.nbsym
        endMin, endMax = len(minPos), len(maxPos)

        ####################################
        # Left bound - mirror nbsym points to the left
        if indmax[0] < indmin[0]:
            if S[0] > S[indmin[0]]:
                lmax = indmax[1 : min(endMax, nbsym + 1)][::-1]
                lmin = indmin[0 : min(endMin, nbsym + 0)][::-1]
                lsym = indmax[0]
            else:
                lmax = indmax[0 : min(endMax, nbsym)][::-1]
                lmin = np.append(indmin[0 : min(endMin, nbsym - 1)][::-1], 0)
                lsym = 0
        else:
            if S[0] < S[indmax[0]]:
                lmax = indmax[0 : min(endMax, nbsym + 0)][::-1]
                lmin = indmin[1 : min(endMin, nbsym + 1)][::-1]
                lsym = indmin[0]
            else:
                lmax = np.append(indmax[0 : min(endMax, nbsym - 1)][::-1], 0)
                lmin = indmin[0 : min(endMin, nbsym)][::-1]
                lsym = 0

        ####################################
        # Right bound - mirror nbsym points to the right
        if indmax[-1] < indmin[-1]:
            if S[-1] < S[indmax[-1]]:
                rmax = indmax[max(endMax - nbsym, 0) :][::-1]
                rmin = indmin[max(endMin - nbsym - 1, 0) : -1][::-1]
                rsym = indmin[-1]
            else:
                rmax = np.append(indmax[max(endMax - nbsym + 1, 0) :], len(S) - 1)[::-1]
                rmin = indmin[max(endMin - nbsym, 0) :][::-1]
                rsym = len(S) - 1
        else:
            if S[-1] > S[indmin[-1]]:
                rmax = indmax[max(endMax - nbsym - 1, 0) : -1][::-1]
                rmin = indmin[max(endMin - nbsym, 0) :][::-1]
                rsym = indmax[-1]
            else:
                rmax = indmax[max(endMax - nbsym, 0) :][::-1]
                rmin = np.append(indmin[max(endMin - nbsym + 1, 0) :], len(S) - 1)[::-1]
                rsym = len(S) - 1

        # In case any array missing
        if not lmin.size:
            lmin = indmin
        if not rmin.size:
            rmin = indmin
        if not lmax.size:
            lmax = indmax
        if not rmax.size:
            rmax = indmax

        # Mirror points
        tlmin = 2 * T[lsym] - T[lmin]
        tlmax = 2 * T[lsym] - T[lmax]
        trmin = 2 * T[rsym] - T[rmin]
        trmax = 2 * T[rsym] - T[rmax]

        # If mirrored points are not outside passed time range.
        if tlmin[0] > T[0] or tlmax[0] > T[0]:
            if lsym == indmax[0]:
                lmax = indmax[0 : min(endMax, nbsym)][::-1]
            else:
                lmin = indmin[0 : min(endMin, nbsym)][::-1]

            if lsym == 0:
                raise Exception("bug")

            lsym = 0
            tlmin = 2 * T[lsym] - T[lmin]
            tlmax = 2 * T[lsym] - T[lmax]

        if trmin[-1] < T[-1] or trmax[-1] < T[-1]:
            if rsym == indmax[-1]:
                rmax = indmax[max(endMax - nbsym, 0) :][::-1]
            else:
                rmin = indmin[max(endMin - nbsym, 0) :][::-1]

            if rsym == len(S) - 1:
                raise Exception("bug")

            rsym = len(S) - 1
            trmin = 2 * T[rsym] - T[rmin]
            trmax = 2 * T[rsym] - T[rmax]

        zlmax = S[lmax]
        zlmin = S[lmin]
        zrmax = S[rmax]
        zrmin = S[rmin]

        tmin = np.append(tlmin, np.append(T[indmin], trmin))
        tmax = np.append(tlmax, np.append(T[indmax], trmax))
        zmin = np.append(zlmin, np.append(S[indmin], zrmin))
        zmax = np.append(zlmax, np.append(S[indmax], zrmax))

        maxExtrema = np.array([tmax, zmax], dtype=self.DTYPE)
        minExtrema = np.array([tmin, zmin], dtype=self.DTYPE)

        # Make double sure, that each extremum is significant
        maxExtrema = np.delete(maxExtrema, np.where(maxExtrema[0, 1:] == maxExtrema[0, :-1]), axis=1)
        minExtrema = np.delete(minExtrema, np.where(minExtrema[0, 1:] == minExtrema[0, :-1]), axis=1)

        return maxExtrema, minExtrema

    def splinePoints(self, T, extrema, splineKind):
        """
        Constructs spline over given points.

        Input:
        ---------
            T: Time array.
            extrema: Position (1st row) and values (2nd row) of points.
            splineKind: Type of spline.

        Output:
        ---------
            T: Position array.
            spline: Spline over the given points.
        """

        kind = splineKind.lower()
        t = T[np.r_[T >= extrema[0, 0]] & np.r_[T <= extrema[0, -1]]]
        if t.dtype != self.DTYPE:
            self.logger.error("t.dtype: " + str(t.dtype))
        if extrema.dtype != self.DTYPE:
            self.logger.error("extrema.dtype: " + str(extrema.dtype))

        if kind == "akima":
            return t, akima(extrema[0], extrema[1], t)

        elif kind == "cubic":
            if extrema.shape[1] > 3:
                return t, interp1d(extrema[0], extrema[1], kind=kind)(t).astype(self.DTYPE)
            else:
                return self.cubicSpline_3points(T, extrema)

        elif kind in ["slinear", "quadratic", "linear"]:
            return T, interp1d(extrema[0], extrema[1], kind=kind)(t).astype(self.DTYPE)

        else:
            raise ValueError("No such interpolation method!")

    def cubicSpline_3points(self, T, extrema):
        """
        Apparently scipy.interpolate.interp1d does not support
        cubic spline for less than 4 points.
        """

        x0, x1, x2 = extrema[0]
        y0, y1, y2 = extrema[1]

        x1x0, x2x1 = x1 - x0, x2 - x1
        y1y0, y2y1 = y1 - y0, y2 - y1
        _x1x0, _x2x1 = 1.0 / x1x0, 1.0 / x2x1

        m11, m12, m13 = 2 * _x1x0, _x1x0, 0
        m21, m22, m23 = _x1x0, 2.0 * (_x1x0 + _x2x1), _x2x1
        m31, m32, m33 = 0, _x2x1, 2.0 * _x2x1

        v1 = 3 * y1y0 * _x1x0 * _x1x0
        v3 = 3 * y2y1 * _x2x1 * _x2x1
        v2 = v1 + v3

        M = np.matrix([[m11, m12, m13], [m21, m22, m23], [m31, m32, m33]])
        v = np.matrix([v1, v2, v3]).T
        k = np.array(np.linalg.inv(M) * v)

        a1 = k[0] * x1x0 - y1y0
        b1 = -k[1] * x1x0 + y1y0
        a2 = k[1] * x2x1 - y2y1
        b2 = -k[2] * x2x1 + y2y1

        t = T[np.r_[T >= x0] & np.r_[T <= x2]]
        t1 = (T[np.r_[T >= x0] & np.r_[T < x1]] - x0) / x1x0
        t2 = (T[np.r_[T >= x1] & np.r_[T <= x2]] - x1) / x2x1
        t11, t22 = 1.0 - t1, 1.0 - t2

        q1 = t11 * y0 + t1 * y1 + t1 * t11 * (a1 * t11 + b1 * t1)
        q2 = t22 * y1 + t2 * y2 + t2 * t22 * (a2 * t22 + b2 * t2)
        q = np.append(q1, q2)

        return t, q.astype(self.DTYPE)

    @classmethod
    def findExtrema(cls, t, s):
        """
        Finds extrema and zero-crossings.

        Input:
        ---------
            S: Signal.
            T: Time array.

        Output:
        ---------
            localMaxPos: Time positions of maxima.
            localMaxVal: Values of signal at localMaxPos positions.
            localMinPos: Time positions of minima.
            localMinVal: Values of signal at localMinPos positions.
            indzer: Indexes of zero crossings.
        """

        # Finds indexes of zero-crossings
        s1, s2 = s[:-1], s[1:]
        indzer = np.nonzero(s1 * s2 < 0)[0]
        if np.any(s == 0):
            iz = np.nonzero(s == 0)[0]
            indz = []
            if np.any(np.diff(iz) == 1):
                zer = s == 0
                dz = np.diff(np.append(np.append(0, zer), 0))
                debz = np.nonzero(dz == 1)[0]
                finz = np.nonzero(dz == -1)[0] - 1
                indz = np.round((debz + finz) / 2)
            else:
                indz = iz

            indzer = np.sort(np.append(indzer, indz))

        # Finds local extrema
        d = np.diff(s)
        d1, d2 = d[:-1], d[1:]
        indmin = np.nonzero(np.r_[d1 * d2 < 0] & np.r_[d1 < 0])[0] + 1
        indmax = np.nonzero(np.r_[d1 * d2 < 0] & np.r_[d1 > 0])[0] + 1

        # When two or more points have the same value
        if np.any(d == 0):
            imax, imin = [], []

            bad = d == 0
            dd = np.diff(np.append(np.append(0, bad), 0))
            debs = np.nonzero(dd == 1)[0]
            fins = np.nonzero(dd == -1)[0]
            if debs[0] == 1:
                if len(debs) > 1:
                    debs, fins = debs[1:], fins[1:]
                else:
                    debs, fins = [], []

            if len(debs) > 0:
                if fins[-1] == len(s) - 1:
                    if len(debs) > 1:
                        debs, fins = debs[:-1], fins[:-1]
                    else:
                        debs, fins = [], []

            lc = len(debs)
            if lc > 0:
                for k in range(lc):
                    if d[debs[k] - 1] > 0:
                        if d[fins[k]] < 0:
                            imax.append(round((fins[k] + debs[k]) / 2.0))
                    else:
                        if d[fins[k]] > 0:
                            imin.append(round((fins[k] + debs[k]) / 2.0))

            if len(imax) > 0:
                indmax = indmax.tolist()
                for x in imax:
                    indmax.append(int(x))
                indmax.sort()

            if len(imin) > 0:
                indmin = indmin.tolist()
                for x in imin:
                    indmin.append(int(x))
                indmin.sort()

        localMaxPos = t[indmax]
        localMaxVal = s[indmax]
        localMinPos = t[indmin]
        localMinVal = s[indmin]

        return localMaxPos, localMaxVal, localMinPos, localMinVal, indzer

    def stop_sifting(self, imf, envMax, envMin, mean, extNo):
        """
        Criterion for stopping sifting process.
        Based on conditions presented in [1].

        [1] G. Rilling, P. Flandrin and P. Goncalves
            "On Empirical Mode Decomposition and its
            algorithms", 2003

        Input:
        ---------
            imf: Current imf.
            envMax: Upper envelope of imf.
            envMin: Bottom envelope of imf.
            mean: Mean of envelopes.
            extNo: Number of extrema.

        Output:
        ---------
            boolean: True if stopping criteria are meet.
        """

        amp = np.abs(envMax - envMin) / 2.0
        sx = np.abs(mean) / amp

        f1 = np.mean(sx > self.stop1) > self.stop3
        f2 = np.any(sx > self.stop2)
        f3 = extNo > 2

        if (not (f1 or f2)) and f3:
            return True
        else:
            return False

    @staticmethod
    def emd(self, S, T=None, maxImf=None):
        """
        Performs Empirical Mode Decomposition on signal S.
        The decomposition is limited to maxImf imf. No limitation as default.
        Returns IMF functions in dic format. IMF = {0:imf0, 1:imf1...}.

        *Note*: First argument `self` should be an instance of EMD class.
        It should be resolved in future versions.

        For example:
        ```
        emd = EMD()
        emd.emd(emd, S, T, maxImf)
        ```

        Input:
        ---------
            self: Instance of EMD class.
            S: Signal.
            T: Positions of signal. If none passed numpy arange is created.
            maxImf: IMF number to which decomposition should be performed.
                    As a default, all IMFs are returned.

        Output:
        ---------
        return IMF, EXT, TIME, ITER, imfNo
            IMF: Signal IMFs in dictionary type. IMF = {0:imf0, 1:imf1...}
            EXT: Number of extrema for each IMF. IMF = {0:ext0, 1:ext1...}
            ITER: Number of iteration for each IMF.
            imfNo: Number of IMFs.
        """

        if T is None:
            T = np.arange(len(S), dtype=S.dtype)
        if maxImf is None:
            maxImf = -1

        # Make sure same types are dealt
        S, T = unify_types(S, T)
        self.DTYPE = S.dtype

        Res = S.astype(self.DTYPE)
        scale = 1.0
        Res, scaledS = Res / scale, S / scale
        imf = np.zeros(len(S), dtype=self.DTYPE)
        imfOld = Res.copy()

        if Res.dtype != self.DTYPE:
            self.logger.error("Res.dtype: " + str(Res.dtype))
        if scaledS.dtype != self.DTYPE:
            self.logger.error("scaledS.dtype: " + str(scaledS.dtype))
        if imf.dtype != self.DTYPE:
            self.logger.error("imf.dtype: " + str(imf.dtype))
        if imfOld.dtype != self.DTYPE:
            self.logger.error("imfOld.dtype: " + str(imfOld.dtype))
        if T.dtype != self.DTYPE:
            self.logger.error("T.dtype: " + str(T.dtype))

        if S.shape != T.shape:
            info = "Time array should be the same size as signal."
            raise ValueError(info)

        # Create arrays
        IMF = {}  # Dic for imfs signals
        EXT = {}  # Dic for number of extrema
        ITER = {}  # Dic for number of iterations
        TIME = {}  # Dic for time of computation
        imfNo = 0
        extNo = 0
        notFinish = True

        while notFinish:
            self.logger.debug("IMF -- " + str(imfNo))

            # ~ Res = scaledS - np.sum([IMF[i] for i in range(imfNo)],axis=0)
            Res -= imf
            imf = Res.copy()
            mean = np.zeros(len(S), dtype=self.DTYPE)

            # Counters
            n = 0  # All iterations for current imf.
            n_h = 0  # counts when |#zero - #ext| <=1

            # Time counter
            timeInit = time.time()
            if self.TIME:
                singleTime = time.time()

            while n < self.MAX_ITERATION:
                n += 1

                if self.TIME:
                    self.logger.info("Execution time: " + str(time.time() - singleTime))
                    singleTime = time.time()
                ext_res = self.findExtrema(T, imf)
                MP, mP = ext_res[0], ext_res[2]
                indzer = ext_res[4]

                extNo = len(mP) + len(MP)
                nzm = len(indzer)

                if extNo > 2:
                    # Plotting. Either into file, or on-screen display.
                    imfOld = imf.copy()
                    imf = imf - self.reduceScale * mean

                    env_ext = self.extractMaxMinSpline(T, imf)
                    maxEnv, minEnv = env_ext[0], env_ext[1]

                    if isinstance(maxEnv, int):
                        notFinish = True
                        break

                    mean = 0.5 * (maxEnv + minEnv)

                    if maxEnv.dtype != self.DTYPE:
                        self.logger.error("maxEnv.dtype: " + str(maxEnv.dtype))
                    if minEnv.dtype != self.DTYPE:
                        self.logger.error("minEnv.dtype: " + str(minEnv.dtype))
                    if imf.dtype != self.DTYPE:
                        self.logger.error("imf.dtype: " + str(imf.dtype))
                    if mean.dtype != self.DTYPE:
                        self.logger.error("mean.dtype: " + str(mean.dtype))

                    # Stop, because of too many iterations
                    if n > self.maxIteration:
                        self.logger.info("TOO MANY ITERATIONS! BREAK!")
                        break

                    # Fix number of iterations
                    if self.FIXE:
                        if n >= self.FIXE + 1:
                            break

                    # Fix number of iterations after number of zero-crossings
                    # and extrema differ at most by one.
                    elif self.FIXE_H:
                        ext_res = self.findExtrema(T, imf)
                        mP, MP, indzer = ext_res[0], ext_res[2], ext_res[4]
                        extNo = len(mP) + len(MP)
                        nzm = len(indzer)

                        if n == 1:
                            continue
                        if abs(extNo - nzm) > 1:
                            n_h = 0
                        else:
                            n_h += 1

                        # STOP
                        if n_h >= self.FIXE_H:
                            break

                    # Stops after default stopping criteria are meet.
                    else:
                        mP, _, MP, _, indzer = self.findExtrema(T, imf)
                        extNo = len(mP) + len(MP)
                        nzm = len(indzer)

                        f1 = self.stop_sifting(imf, maxEnv, minEnv, mean, extNo)
                        f2 = abs(extNo - nzm) < 2

                        # STOP
                        if f1 and f2:
                            break

                else:
                    notFinish = False
                    break

            IMF[imfNo] = imf.copy()
            ITER[imfNo] = n
            EXT[imfNo] = extNo
            TIME[imfNo] = time.time() - timeInit
            imfNo += 1

            if imfNo == maxImf - 1:
                notFinish = False
                break

        # ~ Saving residuum if meaningful
        Res = scaledS - np.sum([IMF[i] for i in range(imfNo)], axis=0)
        if np.sum(np.abs(Res)) > 1e-10:
            IMF[imfNo] = Res
            ITER[imfNo] = 0
            EXT[imfNo] = extNo
            TIME[imfNo] = 0
            imfNo += 1

        for key in list(IMF.keys()):
            IMF[key] *= scale
        return IMF, EXT, ITER, imfNo


###################################################
# Beginning of program

if __name__ == "__main__":
    import pylab as plt

    # Logging options
    logging.basicConfig(level=logging.DEBUG)

    # EMD options
    maxImf = -1
    DTYPE = np.float64

    # Signal options
    N = 1000
    tMin, tMax = 0, 1
    T = np.linspace(tMin, tMax, N, dtype=DTYPE)

    S = 6 * T + np.cos(8 * np.pi**T) + 0.5 * np.cos(40 * np.pi * T)
    S = S.astype(DTYPE)

    # Prepare and run EMD
    emd = EMD()
    emd.FIXE_H = 5
    # ~ emd.FIXE = 10
    emd.nbsym = 2
    emd.splineKind = "cubic"
    emd.DTYPE = DTYPE
    IMF, EXT, ITER, imfNo = emd.emd(S, T, maxImf)

    # Save results (IMFs) into file
    npIMF = np.zeros((imfNo, N), dtype=DTYPE)
    for i in range(imfNo):
        npIMF[i] = IMF[i]

    np.save("imfs", npIMF)

    # Plotting
    # ~ c = np.floor(np.sqrt(imfNo+3))
    # ~ r = np.ceil( (imfNo+3)/c)
    c = np.floor(np.sqrt(imfNo + 1))
    r = np.ceil((imfNo + 1) / c)

    plt.ioff()
    plt.subplot(r, c, 1)
    plt.plot(T, S, "r")
    plt.title("Original signal")
    plt.xlabel("Time [s]")
    plt.ylabel("Amplitude")

    # ~ plt.subplot(r,c,2)
    # ~ plt.plot([EXT[i] for i in range(imfNo)], 'o')
    # ~ plt.ylim(0, max([EXT[i] for i in range(imfNo)])+1)
    # ~ plt.title("Number of extrema")
    # ~
    # ~ plt.subplot(r,c,3)
    # ~ plt.plot([ITER[i] for i in range(imfNo)], 'o')
    # ~ plt.ylim(0, max([ITER[i] for i in range(imfNo)])+1)
    # ~ plt.title("Number of iterations")

    for num in range(imfNo):
        # ~ plt.subplot(r,c,num+4)
        plt.subplot(r, c, num + 2)
        plt.plot(T, IMF[num], "g")
        plt.xlabel("Time")
        plt.ylabel("Amplitude")

        if num == imfNo - 1:
            plt.title("Residue")
        else:
            plt.title("Imf " + str(num))

    plt.tight_layout()
    plt.show()



import numpy as np
from scipy.signal import hilbert

from .compact import filt6, pade6

# Visualisation is an optional module. To minimise installation, `matplotlib` is not added
# by default. Please install extras with `pip install -r requirement-extra.txt`.
try:
    import pylab as plt
except ImportError:
    pass


class Visualisation(object):
    """Simple visualisation helper.

    This class is for quick and simple result visualisation.
    """

    PLOT_WIDTH = 6
    PLOT_HEIGHT_PER_IMF = 1.5

    def __init__(self, emd_instance=None):
        self.emd_instance = emd_instance

        self.imfs = None
        self.residue = None

        if emd_instance is not None:
            self.imfs, self.residue = self.emd_instance.get_imfs_and_residue()

    def _check_imfs(self, imfs, residue, include_residue):
        """Checks for passed imfs and residue."""
        imfs = imfs if imfs is not None else self.imfs
        residue = residue if residue is not None else self.residue

        if imfs is None:
            raise AttributeError("No imfs passed to plot")

        if include_residue and residue is None:
            raise AttributeError("Requested to plot residue but no residue provided")

        return imfs, residue

    def plot_imfs(self, imfs=None, residue=None, t=None, include_residue=True):
        """Plots and shows all IMFs.

        All parameters are optional since the `emd` object could have been passed when instantiating this object.

        The residual is an optional and can be excluded by setting `include_residue=False`.
        """
        imfs, residue = self._check_imfs(imfs, residue, include_residue)

        num_rows, t_length = imfs.shape
        num_rows += include_residue is True

        t = t if t is not None else range(t_length)

        fig, axes = plt.subplots(num_rows, 1, figsize=(self.PLOT_WIDTH, num_rows * self.PLOT_HEIGHT_PER_IMF))

        if num_rows == 1:
            axes = list(axes)

        axes[0].set_title("Time series")

        for num, imf in enumerate(imfs):
            ax = axes[num]
            ax.plot(t, imf)
            ax.set_ylabel("IMF " + str(num + 1))

        if include_residue:
            ax = axes[-1]
            ax.plot(t, residue)
            ax.set_ylabel("Res")

        # Making the layout a bit more pleasant to the eye
        plt.tight_layout()

    def plot_instant_freq(self, t, imfs=None, order=False, alpha=None):
        """Plots and shows instantaneous frequencies for all provided imfs.

        The necessary parameter is `t` which is the time array used to compute the EMD.
        One should pass `imfs` if no `emd` instances is passed when creating the Visualisation object.

        Parameters
        ----------

        order : bool (default: False)
            Represents whether the finite difference scheme is
            low-order (1st order forward scheme) or high-order (6th order
            compact scheme). The default value is False (low-order)

        alpha : float (default: None)
            Filter intensity. Default value is None, which
            is equivalent to `alpha` = 0.5, meaning that no filter is applied.
            The `alpha` values must be in between -0.5 (fully active) and 0.5
            (no filter).
        """
        if alpha is not None:
            assert -0.5 < alpha < 0.5, "`alpha` must be in between -0.5 and 0.5"

        imfs, _ = self._check_imfs(imfs, None, False)
        num_rows = imfs.shape[0]

        imfs_inst_freqs = self._calc_inst_freq(imfs, t, order=order, alpha=alpha)

        fig, axes = plt.subplots(num_rows, 1, figsize=(self.PLOT_WIDTH, num_rows * self.PLOT_HEIGHT_PER_IMF))

        if num_rows == 1:
            axes = fig.axes

        axes[0].set_title("Instantaneous frequency")

        for num, imf_inst_freq in enumerate(imfs_inst_freqs):
            ax = axes[num]
            ax.plot(t, imf_inst_freq)
            ax.set_ylabel("IMF {} [Hz]".format(num + 1))

        # Making the layout a bit more pleasant to the eye
        plt.tight_layout()

    def _calc_inst_phase(self, sig, alpha):
        """Extract analytical signal through the Hilbert Transform."""
        analytic_signal = hilbert(sig)  # Apply Hilbert transform to each row
        if alpha is not None:
            assert -0.5 < alpha < 0.5, "`alpha` must be in between -0.5 and 0.5"
            real_part = np.array([filt6(row.real, alpha) for row in analytic_signal])
            imag_part = np.array([filt6(row.imag, alpha) for row in analytic_signal])
            analytic_signal = real_part + 1j * imag_part
        phase = np.unwrap(np.angle(analytic_signal))  # Compute angle between img and real
        if alpha is not None:
            phase = np.array([filt6(row, alpha) for row in phase])  # Filter phase
        return phase

    def _calc_inst_freq(self, sig, t, order, alpha):
        """Extracts instantaneous frequency through the Hilbert Transform."""
        inst_phase = self._calc_inst_phase(sig, alpha=alpha)
        if order is False:
            inst_freqs = np.diff(inst_phase) / (2 * np.pi * (t[1] - t[0]))
            inst_freqs = np.concatenate((inst_freqs, inst_freqs[:, -1].reshape(inst_freqs[:, -1].shape[0], 1)), axis=1)
        else:
            inst_freqs = [pade6(row, t[1] - t[0]) / (2.0 * np.pi) for row in inst_phase]
        if alpha is None:
            return np.array(inst_freqs)
        else:
            return np.array([filt6(row, alpha) for row in inst_freqs])  # Filter freqs

    def show(self):
        plt.show()


if __name__ == "__main__":
    from .EMD import EMD

    # Simple signal example
    t = np.arange(0, 3, 0.01)
    S = np.sin(13 * t + 0.2 * t**1.4) - np.cos(3 * t)

    emd = EMD()
    emd.emd(S)
    imfs, res = emd.get_imfs_and_residue()

    # Initiate visualisation with emd instance
    vis = Visualisation(emd)

    # Create a plot with all IMFs and residue
    vis.plot_imfs(imfs=imfs, residue=res, t=t, include_residue=True)

    # Create a plot with instantaneous frequency of all IMFs
    vis.plot_instant_freq(t, imfs=imfs)

    # Show both plots
    vis.show()



"""Calculate the statistical Significance of IMFs."""

import logging
import math

import numpy as np
from scipy import stats
from scipy.signal import find_peaks


# helper function: Find mean period of an IMF
def mean_period(data):
    """Return mean-period of signal."""
    peaks = len(find_peaks(data, height=0)[0])
    return len(data) / peaks if peaks > 0 else len(data)


# helper function: find energy of signal/IMF
def energy(data):
    """Return energy of signal."""
    return sum(pow(data, 2))


# helper function: find IMF significance in 'a priori' test
def significance_apriori(energy_density, T, N, alpha):
    """Check a priori significance and Return True if significant else False."""
    k = abs(stats.norm.ppf((1 - alpha) / 2))
    upper_limit = -T + (k * (math.sqrt(2 / N) * math.exp(T / 2)))
    lower_limit = -T - (k * (math.sqrt(2 / N) * math.exp(T / 2)))

    return not (lower_limit <= energy_density <= upper_limit)


# helper function: find significance in 'a posteriori' test
def significance_aposteriori(scaled_energy_density, T, N, alpha):
    """Check a posteriori significance and Return True if significant else False."""
    k = abs(stats.norm.ppf((1 - alpha) / 2))
    upper_limit = -T + (k * (math.sqrt(2 / N) * math.exp(T / 2)))
    return not (scaled_energy_density <= upper_limit)


def whitenoise_check(IMFs: np.ndarray, test_name: str = "aposteriori", rescaling_imf: int = 1, alpha: float = 0.95):
    """Whitenoise statistical significance test.

    Performs whitenoise test as described by Wu & Huang [Wu2004]_.

    References
    ----------
    .. [Wu2004] Zhaohua Wu, and Norden E. Huang. "A Study of the Characteristics of White Noise Using the
       Empirical Mode Decomposition Method." Proceedings: Mathematical, Physical and Engineering
       Sciences, vol. 460, no. 2046, The Royal Society, 2004, pp. 1597–611, http://www.jstor.org/stable/4143111.

    Parameters
    ----------
    IMFs: np.ndarray
        (Required) numpy array containing IMFs computed from a normalized signal
    test_name: str
        (Optional) Test type. Supported values: 'apriori', 'aposteriori'. (default 'aposteriori')
    rescaling_imf: int
        (Optional) ith IMF of the signal used in rescaling for 'a posteriori' test. (default 1)
    alpha: float
        (Optional) The percentiles at which the test is to be performed; 0 < alpha < 1; (default 0.95)

    Returns
    -------
    Optional dictionary
        Returns dictionary with keys and values as IMFs' number and test result, respetively,
        Test results can be either 0 (fail) or 1 (pass).
        In case of problems with the input imfs, e.g. NaN values or no imfs, we return None.

    Examples
    --------
    >>> import numpy as np
    >>> from PyEMD import EMD
    >>> from PyEMD.checks import whitenoise_check
    >>> T = np.linspace(0, 1, 100)
    >>> S = np.sin(2*2*np.pi*T)
    >>> emd = EMD()
    >>> imfs = emd(S)
    >>> significant_imfs = whitenoise_check(imfs, test_name='apriori')
    >>> significant_imfs
    {1: 1, 2: 1}
    >>> type(significant_imfs)
    <class 'dict'>

    """
    assert isinstance(IMFs, np.ndarray), "Invalid Data type, Pass a numpy.ndarray containing IMFs"
    # check if IMFs are empty or not
    if len(IMFs) == 0 or len(IMFs[0]) == 0:
        logging.getLogger("PyEMD").warning("Detected empty input. Skipping check.")
        return None

    assert isinstance(alpha, float), "Invalid Data type for alpha, pass a float value between (0,1)"
    assert 0 < alpha < 1, "alpha value should be in between (0,1)"
    assert test_name in ("apriori", "aposteriori"), "Invalid test type"
    assert isinstance(rescaling_imf, int), "Invalid data type for rescaling_imf, pass a int value"
    assert 0 < rescaling_imf <= len(IMFs), "Invalid rescaling IMF"

    if np.any(np.isnan(IMFs)):
        # Return None if input has NaN
        logging.getLogger("PyEMD").warning("Detected NaN values during whitenoise check. Skipping check.")
        return None

    N = len(IMFs[0])
    output = {}
    if test_name == "apriori":
        for idx, imf in enumerate(IMFs):
            log_T = math.log(mean_period(imf))
            energy_density = math.log(energy(imf) / N)
            sig_priori = significance_apriori(energy_density, log_T, N, alpha)

            output[idx + 1] = int(sig_priori)

    elif test_name == "aposteriori":
        scaling_imf_mean_period = math.log(mean_period(IMFs[rescaling_imf - 1]))
        scaling_imf_energy_density = math.log(energy(IMFs[rescaling_imf - 1]) / N)

        k = abs(stats.norm.ppf((1 - alpha) / 2))
        up_limit = -scaling_imf_mean_period + (k * math.sqrt(2 / N) * math.exp(scaling_imf_mean_period) / 2)

        scaling_factor = math.exp(up_limit)

        for idx, imf in enumerate(IMFs):
            log_T = math.log(mean_period(imf))
            if idx != rescaling_imf - 1:
                scaled_energy_density = math.log((energy(imf) / N) / scaling_factor)
            else:
                scaled_energy_density = math.log(scaling_factor)

            sig_aposteriori = significance_aposteriori(scaled_energy_density, log_T, N, alpha)

            output[idx + 1] = int(sig_aposteriori)

    else:
        raise AssertionError("Only 'apriori' and 'aposteriori' are allowed")

    return output



# -----------------------------------------------------------------------------------------------------
# WaLSAtools: Wave analysis tools
# Copyright (C) 2025 WaLSA Team - Shahin Jafarzadeh
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Note: If you use WaLSAtools for research, please consider citing:
# Jafarzadeh, S., Jess, D. B., Stangalini, M. et al. 2025, Nature Reviews Methods Primers, in press.
# -----------------------------------------------------------------------------------------------------

from .analysis_modules.WaLSA_detrend_apod import WaLSA_detrend_apod
from .analysis_modules.WaLSA_speclizer import WaLSA_speclizer
from .analysis_modules.WaLSA_k_omega import WaLSA_k_omega
from .analysis_modules.WaLSA_pod import WaLSA_pod
from .analysis_modules.WaLSA_cross_spectra import WaLSA_cross_spectra
from .analysis_modules.WaLSA_io import WaLSA_save_pdf, WaLSA_histo_opt
from .analysis_modules.WaLSA_plot_k_omega import WaLSA_plot_k_omega
from .WaLSAtools import WaLSAtools

__all__ = ['WaLSAtools', 'WaLSA_detrend_apod', 'WaLSA_speclizer',
           'WaLSA_save_pdf', 'WaLSA_histo_opt', 'WaLSA_plot_k_omega', 
           'WaLSA_k_omega', 'WaLSA_pod', 'WaLSA_cross_spectra']



# -----------------------------------------------------------------------------------------------------
# WaLSAtools - Wave analysis tools
# Copyright (C) 2025 WaLSA Team - Shahin Jafarzadeh et al.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Note: If you use WaLSAtools for research, please consider citing:
# Jafarzadeh, S., Jess, D. B., Stangalini, M. et al. 2025, Nature Reviews Methods Primers, in press.
# -----------------------------------------------------------------------------------------------------

from IPython.display import HTML, display
import shutil

single_series_parameters = {
    'fft': {
        'return_values': 'power, frequency, significance, amplitude',
        'parameters': {
            'signal': {'type': 'array', 'description': 'The input signal (1D).'},
            'time': {'type': 'array', 'description': 'The time array corresponding to the signal.'},
            'siglevel': {'type': 'float', 'description': 'Significance level for the confidence intervals. Default: 0.95.'},
            'nperm': {'type': 'int', 'description': 'Number of permutations for significance testing. Default: 1000.'},
            'nosignificance': {'type': 'bool', 'description': 'If True, skip significance calculation. Default: False.'},
            'apod': {'type': 'float', 'description': 'Extent of apodization edges (of a Tukey window). Default: 0.1.'},
            'pxdetrend': {'type': 'int', 'description': 'Subtract linear trend with time per pixel. Options: 1 (simple) or 2 (advanced). Default: 2.'},
            'polyfit': {'type': 'int', 'description': 'Degree of polynomial fit for detrending the data. If set, a polynomial fit (instead of linear) is applied. Default: None'},
            'meantemporal': {'type': 'bool', 'description': 'If True, apply simple temporal detrending by subtracting the mean signal from the data, skipping fitting procedures. Default: False.'},
            'meandetrend': {'type': 'bool', 'description': 'If True, subtract the linear trend with time for the image means (spatial detrending). Default: False.'},
            'recon': {'type': 'bool', 'description': 'If True, perform Fourier reconstruction of the input time series. This does not preserve amplitudes but is useful for examining frequencies far from the low-frequency range. Default: False.'},
            'resample_original': {'type': 'bool', 'description': 'If True, and if recon is set to True, approximate values close to the original are returned for comparison. Default: False.'},
            'nodetrendapod': {'type': 'bool', 'description': 'If True, neither detrending nor apodization is performed. Default: False.'},
            'amplitude': {'type': 'bool', 'description': ' If True, return the amplitudes of the Fourier transform. Default: False.'},
            'silent': {'type': 'bool', 'description': 'If True, suppress print statements. Default: False.'}
        }
    },
    'wavelet': {
        'return_values': 'power, period, significance, coi, gws_power, gws_significance, rgws_power',
        'parameters': {
            'signal': {'type': 'array', 'description': 'The input signal (1D).'},
            'time': {'type': 'array', 'description': 'The time array corresponding to the signal.'},
            'siglevel': {'type': 'float', 'description': 'Significance level for the confidence intervals. Default: 0.95.'},
            'nperm': {'type': 'int', 'description': 'Number of permutations for significance testing. Default: 1000.'},
            'mother': {'type': 'str', 'description': 'The mother wavelet function to use. Default: "morlet".'},
            'GWS': {'type': 'bool', 'description': 'If True, calculate the Global Wavelet Spectrum. Default: False.'},
            'RGWS': {'type': 'bool', 'description': 'If True, calculate the Refined Global Wavelet Spectrum (time-integrated power, excluding COI and insignificant areas). Default: False.'},
            'dj': {'type': 'float', 'description': 'Scale spacing. Smaller values result in better scale resolution but slower calculations. Default: 0.025.'},
            's0': {'type': 'float', 'description': 'Initial (smallest) scale of the wavelet. Default: 2 * dt.'},
            'J': {'type': 'int', 'description': 'Number of scales minus one. Scales range from s0 up to s0 * 2**(J * dj), giving a total of (J + 1) scales. Default: (log2(N * dt / s0)) / dj.'},
            'lag1': {'type': 'float', 'description': 'Lag-1 autocorrelation. Default: 0.0.'},
            'apod': {'type': 'float', 'description': 'Extent of apodization edges (of a Tukey window). Default: 0.1.'},
            'pxdetrend': {'type': 'int', 'description': 'Subtract linear trend with time per pixel. Options: 1 (simple) or 2 (advanced). Default: 2.'},
            'polyfit': {'type': 'int', 'description': 'Degree of polynomial fit for detrending the data. If set, a polynomial fit (instead of linear) is applied. Default: None'},
            'meantemporal': {'type': 'bool', 'description': 'If True, apply simple temporal detrending by subtracting the mean signal from the data, skipping fitting procedures. Default: False.'},
            'meandetrend': {'type': 'bool', 'description': 'If True, subtract the linear trend with time for the image means (spatial detrending). Default: False.'},
            'recon': {'type': 'bool', 'description': 'If True, perform Fourier reconstruction of the input time series. This does not preserve amplitudes but is useful for examining frequencies far from the low-frequency range. Default: False.'},
            'resample_original': {'type': 'bool', 'description': 'If True, and if recon is set to True, approximate values close to the original are returned for comparison. Default: False.'},
            'nodetrendapod': {'type': 'bool', 'description': 'If True, neither detrending nor apodization is performed. Default: False.'},
            'silent': {'type': 'bool', 'description': 'If True, suppress print statements. Default: False.'}
        }
    },
    'lombscargle': {
        'return_values': 'power, frequency, significance',
        'parameters': {
            'signal': {'type': 'array', 'description': 'The input signal (1D).'},
            'time': {'type': 'array', 'description': 'The time array corresponding to the signal.'},
            'siglevel': {'type': 'float', 'description': 'Significance level for the confidence intervals. Default: 0.95.'},
            'nperm': {'type': 'int', 'description': 'Number of permutations for significance testing. Default: 1000.'},
            'dy': {'type': 'array', 'description': 'Errors or observational uncertainties associated with the time series.'},
            'fit_mean': {'type': 'bool', 'description': 'If True, include a constant offset as part of the model at each frequency. This improves accuracy, especially for incomplete phase coverage.'},
            'center_data': {'type': 'bool', 'description': 'If True, pre-center the data by subtracting the weighted mean of the input data. This is especially important if fit_mean=False.'},
            'nterms': {'type': 'int', 'description': 'Number of terms to use in the Fourier fit. Default: 1.'},
            'normalization': {'type': 'str', 'description': 'The normalization method for the periodogram. Options: "standard", "model", "log", "psd". Default: "standard".'},
            'nosignificance': {'type': 'bool', 'description': 'If True, skip significance calculation. Default: False.'},
            'apod': {'type': 'float', 'description': 'Extent of apodization edges (of a Tukey window). Default: 0.1.'},
            'pxdetrend': {'type': 'int', 'description': 'Subtract linear trend with time per pixel. Options: 1 (simple) or 2 (advanced). Default: 2.'},
            'polyfit': {'type': 'int', 'description': 'Degree of polynomial fit for detrending the data. If set, a polynomial fit (instead of linear) is applied. Default: None'},
            'meantemporal': {'type': 'bool', 'description': 'If True, apply simple temporal detrending by subtracting the mean signal from the data, skipping fitting procedures. Default: False.'},
            'meandetrend': {'type': 'bool', 'description': 'If True, subtract the linear trend with time for the image means (spatial detrending). Default: False.'},
            'recon': {'type': 'bool', 'description': 'If True, perform Fourier reconstruction of the input time series. This does not preserve amplitudes but is useful for examining frequencies far from the low-frequency range. Default: False.'},
            'resample_original': {'type': 'bool', 'description': 'If True, and if recon is set to True, approximate values close to the original are returned for comparison. Default: False.'},
            'nodetrendapod': {'type': 'bool', 'description': 'If True, neither detrending nor apodization is performed. Default: False.'},
            'silent': {'type': 'bool', 'description': 'If True, suppress print statements. Default: False.'}
        }
    },
    'welch': {
        'return_values': 'power, frequency, significance',
        'parameters': {
            'signal': {'type': 'array', 'description': 'The 1D time series signal.'},
            'time': {'type': 'array', 'description': 'The time array corresponding to the signal.'},
            'nperseg': {'type': 'int', 'description': 'Length of each segment for analysis. Default: 256.'},
            'noverlap': {'type': 'int', 'description': 'Number of points to overlap between segments. Default: 128.'},
            'window': {'type': 'str', 'description': 'Type of window function used in the Welch method. Default: "hann".'},
            'siglevel': {'type': 'float', 'description': 'Significance level for confidence intervals. Default: 0.95.'},
            'nperm': {'type': 'int', 'description': 'Number of permutations for significance testing. Default: 1000.'},
            'silent': {'type': 'bool', 'description': 'If True, suppress print statements. Default: False.'}
        }
    },
    'emd': {
        'return_values': 'HHT_power, HHT_significance, HHT_frequency, psd_spectra, psd_significance, IMFs, IMF_significance, instantaneous_frequency',
        'parameters': {
            'signal': {'type': 'array', 'description': 'The input signal (1D).'},
            'time': {'type': 'array', 'description': 'The time array of the signal.'},
            'siglevel': {'type': 'float', 'description': 'Significance level for the confidence intervals. Default: 0.95.'},
            'nperm': {'type': 'int', 'description': 'Number of permutations for significance testing. Default: 1000.'},
            'EEMD': {'type': 'bool', 'description': 'If True, use Ensemble Empirical Mode Decomposition (EEMD) instead of Empirical Mode Decomposition (EMD). Default: False.'},
            'Welch_psd': {'type': 'bool', 'description': 'If True, calculate Welch PSD spectra instead of FFT PSD spectra (for the psd_spectra and psd_confidence_levels). Default: False.'},
            'significant_imfs': {'type': 'bool', 'description': 'If True, only significant IMFs and their associated parameters are retured. Default: False.'},
            'apod': {'type': 'float', 'description': 'Extent of apodization edges (of a Tukey window). Default: 0.1.'},
            'pxdetrend': {'type': 'int', 'description': 'Subtract linear trend with time per pixel. Options: 1 (simple) or 2 (advanced). Default: 2.'},
            'polyfit': {'type': 'int', 'description': 'Degree of polynomial fit for detrending the data. If set, a polynomial fit (instead of linear) is applied. Default: None.'},
            'meantemporal': {'type': 'bool', 'description': 'If True, apply simple temporal detrending by subtracting the mean signal from the data, skipping fitting procedures. Default: False.'},
            'meandetrend': {'type': 'bool', 'description': 'If True, subtract the linear trend with time for the image means (spatial detrending). Default: False.'},
            'recon': {'type': 'bool', 'description': 'If True, perform Fourier reconstruction of the input time series. This does not preserve amplitudes but is useful for examining frequencies far from the low-frequency range. Default: False.'},
            'resample_original': {'type': 'bool', 'description': 'If True, and if recon is set to True, approximate values close to the original are returned for comparison. Default: False.'},
            'nodetrendapod': {'type': 'bool', 'description': 'If True, neither detrending nor apodization is performed. Default: False.'},
            'silent': {'type': 'bool', 'description': 'If True, suppress print statements. Default: False.'}
        }
    },
    'k-omega': {
        'return_values': 'power, wavenumber, frequency, filtered_cube, spatial_fft_map, torus_map, spatial_fft_filtered_map, temporal_fft, temporal_filter, temporal_frequencies, spatial_frequencies',
        'parameters': {
            'signal': {'type': 'array', 'description': 'Input datacube, normally in the form of [x, y, t] or [t, x, y]. Note that the input datacube must have identical x and y dimensions. If not, the datacube will be cropped accordingly.'},
            'time': {'type': 'array', 'description': 'Time array corresponding to the input datacube.'},
            'pixelsize': {'type': 'float', 'description': 'Spatial sampling of the input datacube. If not given, it is plotted in units of "pixel".'},
            'filtering': {'type': 'bool', 'description': 'If True, filtering is applied, and the filtered datacube (filtered_cube) is returned. Otherwise, None is returned. Default: False.'},
            'f1': {'type': 'float', 'description': 'Optional lower (temporal) frequency to filter, in Hz.'},
            'f2': {'type': 'float', 'description': 'Optional upper (temporal) frequency to filter, in Hz.'},
            'k1': {'type': 'float', 'description': 'Optional lower (spatial) wavenumber to filter, in units of pixelsize^-1 (k = (2 * π) / wavelength).'},
            'k2': {'type': 'float', 'description': 'Optional upper (spatial) wavenumber to filter, in units of pixelsize^-1.'},
            'spatial_torus': {'type': 'bool', 'description': 'If True, makes the annulus used for spatial filtering have a Gaussian-shaped profile, useful for preventing aliasing. Default: True.'},
            'temporal_torus': {'type': 'bool', 'description': 'If True, makes the temporal filter have a Gaussian-shaped profile, useful for preventing aliasing. Default: True.'},
            'no_spatial_filt': {'type': 'bool', 'description': 'If True, ensures no spatial filtering is performed on the dataset (i.e., only temporal filtering is applied).'},
            'no_temporal_filt': {'type': 'bool', 'description': 'If True, ensures no temporal filtering is performed on the dataset (i.e., only spatial filtering is applied).'},
            'silent': {'type': 'bool', 'description': 'If True, suppresses the k-ω diagram plot.'},
            'smooth': {'type': 'bool', 'description': 'If True, power is smoothed. Default: True.'},
            'mode': {'type': 'int', 'description': 'Output power mode: 0 = log10(power) (default), 1 = linear power, 2 = sqrt(power) = amplitude.'},
            'processing_maps': {'type': 'bool', 'description': 'If True, the function returns the processing maps (spatial_fft_map, torus_map, spatial_fft_filtered_map, temporal_fft, temporal_filter, temporal_frequencies, spatial_frequencies). Otherwise, they are all returned as None. Default: False.'}
        }
    },
    'pod': {
        'return_values': 'pod_results',
        'parameters': {
            'signal': {'type': 'array', 'description': '3D data cube with shape (time, x, y) or similar.'},
            'time': {'type': 'array', 'description': '1D array representing the time points for each time step in the data.'},
            'num_modes': {'type': 'int, optional', 'description': 'Number of top modes to compute. Default is None (all modes).'},
            'num_top_frequencies': {'type': 'int, optional', 'description': 'Number of top frequencies to consider. Default is None (all frequencies).'},
            'top_frequencies': {'type': 'list, optional', 'description': 'List of top frequencies to consider. Default is None.'},
            'num_cumulative_modes': {'type': 'int, optional', 'description': 'Number of cumulative modes to consider. Default is None (all modes).'},
            'welch_nperseg': {'type': 'int, optional', 'description': "Number of samples per segment for Welch's method. Default is 150."},
            'welch_noverlap': {'type': 'int, optional', 'description': "Number of overlapping samples for Welch's method. Default is 25."},
            'welch_nfft': {'type': 'int, optional', 'description': 'Number of points for the FFT. Default is 2^14.'},
            'welch_fs': {'type': 'int, optional', 'description': 'Sampling frequency for the data. Default is 2.'},
            'nperm': {'type': 'int, optional', 'description': 'Number of permutations for significance testing. Default is 1000.'},
            'siglevel': {'type': 'float, optional', 'description': 'Significance level for the Welch spectrum. Default is 0.95.'},
            'timestep_to_reconstruct': {'type': 'int, optional', 'description': 'Timestep of the datacube to reconstruct using the top modes. Default is 0.'},
            'num_modes_reconstruct': {'type': 'int, optional', 'description': 'Number of modes to use for reconstruction. Default is None (all modes).'},
            'reconstruct_all': {'type': 'bool, optional', 'description': 'If True, reconstruct the entire time series using the top modes. Default is False.'},
            'spod': {'type': 'bool, optional', 'description': 'If True, perform Spectral Proper Orthogonal Decomposition (SPOD) analysis. Default is False.'},
            'spod_filter_size': {'type': 'int, optional', 'description': 'Filter size for SPOD analysis. Default is None.'},
            'spod_num_modes': {'type': 'int, optional', 'description': 'Number of SPOD modes to compute. Default is None.'},
            'print_results': {'type': 'bool, optional', 'description': 'If True, print a summary of results. Default is True.'}
        }
    },
    'dominantfreq': {
        'return_values': 'power, frequency, significance',
        'parameters': {
            'signal': 'Input signal array (1D or 2D).',
            'time': 'Time array corresponding to the signal.',
            'method': 'Analysis method (e.g., fft, wavelet, etc.).',
            'kwargs': 'Additional optional parameters for customization.'
        }
    }
}

cross_correlation_parameters = {
    'wavelet': {
        'return_values': 'cross_power, cross_period, cross_sig, cross_coi, coherence, coh_period, coh_sig, coh_coi, phase_angle',
        'parameters': {
            'data1': {'type': 'array', 'description': 'The first 1D time series signal.'},
            'data2': {'type': 'array', 'description': 'The second 1D time series signal.'},
            'time': {'type': 'array', 'description': 'The time array corresponding to the signals.'},
            'siglevel': {'type': 'float', 'description': 'Significance level for the confidence intervals. Default: 0.95.'},
            'nperm': {'type': 'int', 'description': 'Number of permutations for significance testing. Default: 1000.'},
            'mother': {'type': 'str', 'description': 'The mother wavelet function to use. Default: "morlet".'},
            'GWS': {'type': 'bool', 'description': 'If True, calculate the Global Wavelet Spectrum. Default: False.'},
            'RGWS': {'type': 'bool', 'description': 'If True, calculate the Refined Global Wavelet Spectrum (time-integrated power, excluding COI and insignificant areas). Default: False.'},
            'dj': {'type': 'float', 'description': 'Scale spacing. Smaller values result in better scale resolution but slower calculations. Default: 0.025.'},
            's0': {'type': 'float', 'description': 'Initial (smallest) scale of the wavelet. Default: 2 * dt.'},
            'J': {'type': 'int', 'description': 'Number of scales minus one. Scales range from s0 up to s0 * 2**(J * dj), giving a total of (J + 1) scales. Default: (log2(N * dt / s0)) / dj.'},
            'lag1': {'type': 'float', 'description': 'Lag-1 autocorrelation. Default: 0.0.'},
            'apod': {'type': 'float', 'description': 'Extent of apodization edges (of a Tukey window). Default: 0.1.'},
            'pxdetrend': {'type': 'int', 'description': 'Subtract linear trend with time per pixel. Options: 1 (simple) or 2 (advanced). Default: 2.'},
            'polyfit': {'type': 'int', 'description': 'Degree of polynomial fit for detrending the data. If set, a polynomial fit (instead of linear) is applied. Default: None.'},
            'meantemporal': {'type': 'bool', 'description': 'If True, apply simple temporal detrending by subtracting the mean signal from the data, skipping fitting procedures. Default: False.'},
            'meandetrend': {'type': 'bool', 'description': 'If True, subtract the linear trend with time for the image means (spatial detrending). Default: False.'},
            'recon': {'type': 'bool', 'description': 'If True, perform Fourier reconstruction of the input time series. This does not preserve amplitudes but is useful for examining frequencies far from the low-frequency range. Default: False.'},
            'resample_original': {'type': 'bool', 'description': 'If True, and if recon is set to True, approximate values close to the original are returned for comparison. Default: False.'},
            'nodetrendapod': {'type': 'bool', 'description': 'If True, neither detrending nor apodization is performed. Default: False.'},
            'silent': {'type': 'bool', 'description': 'If True, suppress print statements. Default: False.'}
        }
    },
    'welch': {
        'return_values': 'frequency, cospectrum, phase_angle, power_data1, power_data2, frequency_coherence, coherence',
        'parameters': {
            'data1': {'type': 'array', 'description': 'The first 1D time series signal.'},
            'data2': {'type': 'array', 'description': 'The second 1D time series signal.'},
            'time': {'type': 'array', 'description': 'The time array corresponding to the signals.'},
            'nperseg': {'type': 'int', 'description': 'Length of each segment for analysis. Default: 256.'},
            'noverlap': {'type': 'int', 'description': 'Number of points to overlap between segments. Default: 128.'},
            'window': {'type': 'str', 'description': 'Type of window function used in the Welch method. Default: "hann".'},
            'siglevel': {'type': 'float', 'description': 'Significance level for confidence intervals. Default: 0.95.'},
            'nperm': {'type': 'int', 'description': 'Number of permutations for significance testing. Default: 1000.'},
            'silent': {'type': 'bool', 'description': 'If True, suppress print statements. Default: False.'}
        }
    }
}


# For the Jupyter-based interactive function
def display_parameters_html(method_name, category):
    """Display parameters as an HTML table (for Jupyter Notebooks)."""
    if category == 'b: Cross-correlation between two time series':
        parameter_definitions = cross_correlation_parameters
    else:
        parameter_definitions = single_series_parameters
    parameters = parameter_definitions.get(method_name, {}).get('parameters')
    if not parameters:
        display(HTML(f"<p><strong>No parameters available for method: {method_name}</strong></p>"))
        return
			
    table = '<table style="border-collapse: collapse; border: 1px solid #222; width: calc(100% - 30px); box-sizing: border-box; margin-left: 30px; table-layout: auto;">'
    table += '<tr style="background-color: #fff;"><th colspan="3" style="text-align: left; color: #000; font-size: 110%;">Parameters (**kwargs)</th></tr>'
    table += '<tr style="background-color: #222;">'
    table += '<th style="color: #fff; border-right: 1px solid #ccc; text-align: left; white-space: nowrap;"><strong>Parameter</strong></th>'
    table += '<th style="color: #fff; border-right: 1px solid #ccc; text-align: left; white-space: nowrap;"><strong>Type</strong></th>'
    table += '<th style="color: #fff; text-align: left; width: 100%;"><strong>Description</strong></th>'
    table += '</tr>'
    for param, info in parameters.items():
        param_type = info['type']
        description = info['description']
        table += f'<tr>'
        table += f'<td style="border: 1px solid #222; padding: 8px; text-align: left; white-space: nowrap;">{param}</td>'
        table += f'<td style="border: 1px solid #222; padding: 8px; text-align: left; white-space: nowrap;">{param_type}</td>'
        table += f'<td style="border: 1px solid #222; padding: 8px; text-align: left;">{description}</td>'
        table += '</tr>'
    table += '</table>'
    display(HTML(table))

# For the Terminal-based interactive function
def display_parameters_text(method_name, category):
    """Display parameters as a plain text table for the terminal."""
    if category == 'b':
        parameter_definitions = cross_correlation_parameters
    else:
        parameter_definitions = single_series_parameters    
    parameters = parameter_definitions.get(method_name, {}).get('parameters')
    if not parameters:
        print(f"No parameters available for method: {method_name}")
        return
    
    # Get terminal width
    terminal_width = shutil.get_terminal_size().columns
    # Calculate 90% of the width (and ensure it's an integer)
    line_width = int(terminal_width * 0.90)

    print('    '+'-' * line_width)
    print('    | Parameters (**kwargs):')
    print('    '+'-' * line_width)
    print(f'    | {"Parameter":<15} | {"Type":<8} | {"Description":<30} ')
    print('    '+'-' * line_width)
    for param, info in parameters.items():
        param_type = info['type']
        description = info['description']
        print(f'    | {param:<15} | {param_type:<8}| {description:<30} ')
    print('    '+'-' * line_width)



# -----------------------------------------------------------------------------------------------------
# WaLSAtools - Wave analysis tools
# Copyright (C) 2025 WaLSA Team - Shahin Jafarzadeh et al.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Note: If you use WaLSAtools for research, please consider citing:
# Jafarzadeh, S., Jess, D. B., Stangalini, M. et al. 2025, Nature Reviews Methods Primers, in press.
# -----------------------------------------------------------------------------------------------------

import numpy as np # type: ignore

def WaLSA_wavelet_confidence(ps_perm, siglevel=None):
    """
    Find the confidence levels (significance levels) for the given wavelet power spectrum permutations.

    Parameters:
    -----------
    ps_perm : np.ndarray
        2D array where each row represents a power spectrum, and each column represents a permutation.
    siglevel : float
        e.g., 0.95 for 95% confidence level (significance at 5%).

    Returns:
    --------
    signif : np.ndarray
        Significance levels for each frequency.
    """
    nnff , nntt, nnperm = ps_perm.shape

    signif = np.zeros((nnff, nntt))

    for iv in range(nnff):
        for it in range(nntt):
            signif[iv, it] = np.percentile(ps_perm[iv, it, :], 100 * siglevel) # e.g., 95% percentile (for 95% consideence level)

    return signif



#!/usr/bin/python
# coding: UTF-8
#
# Author:   Dawid Laszuk
# Contact:  https://github.com/laszukdawid/PyEMD/issues
#
# Feel free to contact for any information.

import logging

import numpy as np
from scipy.interpolate import Rbf

try:
    from skimage.morphology import reconstruction
except (ImportError, ModuleNotFoundError):
    raise ImportError(
        "EMD2D and BEMD are not supported. Feel free to play around and improve them. "
        + "Required dependencies are in `requirements-extra`."
    )


class BEMD:
    """
    **Bidimensional Empirical Mode Decomposition**

    **Important**: This class intends to be undocumented until it's actually properly tested
    and proven to work. An attempt to replicate findings in the paper cited below has failed.
    This method is only included in the package because someone asked for it, and I'm hoping
    that one day someone else will come and *fix it*. Until then, USE AT YOUR OWN RISK.

    The guess why the decomposition doesn't work is that it's difficult to extrapolate image
    far away from extrema. Not even mirroring helps in this case.

    Method decomposition 2D arrays like gray-scale images into 2D representations of
    Intrinsic Mode Functions (IMFs).

    The algorithm is based on Nunes et. al. [Nunes2003]_ work.

    .. [Nunes2003] J.-C. Nunes, Y. Bouaoune, E. Delechelle, O. Niang, P. Bunel.,
       "Image analysis by bidimensional empirical mode decomposition. Image and Vision Computing",
       Elsevier, 2003, 21 (12), pp.1019-1026.
    """

    logger = logging.getLogger(__name__)

    def __init__(self):
        # ProtoIMF related
        self.mse_thr = 0.01
        self.mean_thr = 0.01

        self.FIXE = 1  # Single iteration by default, otherwise results are terrible
        self.FIXE_H = 0
        self.MAX_ITERATION = 5

    def __call__(self, image, max_imf=-1):
        return self.bemd(image, max_imf=max_imf)

    def extract_max_min_spline(self, image, min_peaks_pos, max_peaks_pos):
        """Calculates top and bottom envelopes for image.

        Parameters
        ----------
        image : numpy 2D array

        Returns
        -------
        min_env : numpy 2D array
            Bottom envelope in form of an image.
        max_env : numpy 2D array
            Top envelope in form of an image.
        """
        xi, yi = np.meshgrid(np.arange(image.shape[0]), np.arange(image.shape[1]))
        min_val = np.array([image[x, y] for x, y in zip(*min_peaks_pos)])
        max_val = np.array([image[x, y] for x, y in zip(*max_peaks_pos)])
        min_env = self.spline_points(min_peaks_pos[0], min_peaks_pos[1], min_val, xi, yi)
        max_env = self.spline_points(max_peaks_pos[0], max_peaks_pos[1], max_val, xi, yi)
        return min_env, max_env

    @classmethod
    def spline_points(cls, X, Y, Z, xi, yi):
        """Creates a spline for given set of points.

        Uses Radial-basis function to extrapolate surfaces. It's not the best but gives something.
        Grid data algorithm didn't work.
        """
        spline = Rbf(X, Y, Z, function="cubic")
        return spline(xi, yi)

    @classmethod
    def find_extrema_positions(cls, image):
        """
        Finds extrema, both minima and maxima, based on morphological reconstruction.
        Returns extrema where the first and second elements are x and y positions, respectively.

        Parameters
        ----------
        image : numpy 2D array
            Monochromatic image or any 2D array.

        Returns
        -------
        min_peaks_pos : numpy array
            Minima positions.
        max_peaks_pos : numpy array
            Maxima positions.
        """
        min_peaks_pos = BEMD.extract_minima_positions(image)
        max_peaks_pos = BEMD.extract_maxima_positions(image)
        return min_peaks_pos, max_peaks_pos

    @classmethod
    def extract_minima_positions(cls, image):
        return BEMD.extract_maxima_positions(-image)

    @classmethod
    def extract_maxima_positions(cls, image):
        seed_min = image - 1
        dilated = reconstruction(seed_min, image, method="dilation")
        cleaned_image = image - dilated
        return np.where(cleaned_image > 0)[::-1]

    @classmethod
    def end_condition(cls, image, IMFs):
        """Determines whether decomposition should be stopped.

        Parameters
        ----------
        image : numpy 2D array
            Input image which is decomposed.
        IMFs : numpy 3D array
            Array for which first dimensions relates to respective IMF,
            i.e. (numIMFs, imageX, imageY).
        """
        rec = np.sum(IMFs, axis=0)

        # If reconstruction is perfect, no need for more tests
        if np.allclose(image, rec):
            return True

        return False

    def check_proto_imf(self, proto_imf, proto_imf_prev, mean_env):
        """Check whether passed (proto) IMF is actual IMF.
        Current condition is solely based on checking whether the mean is below threshold.

        Parameters
        ----------
        proto_imf : numpy 2D array
            Current iteration of proto IMF.
        proto_imf_prev : numpy 2D array
            Previous iteration of proto IMF.
        mean_env : numpy 2D array
            Local mean computed from top and bottom envelopes.

        Returns
        -------
        boolean
            Whether current proto IMF is actual IMF.
        """
        # TODO: Sifting is very sensitive and subtracting const val can often flip
        #      maxima with minima in decomposition and thus repeating above/below
        #      behaviour. For now, mean_env is checked whether close to zero excluding
        #      its offset.
        if np.all(np.abs(mean_env - mean_env.mean()) < self.mean_thr):
            # if np.all(np.abs(mean_env)<self.mean_thr):
            return True

        # If very little change with sifting
        if np.allclose(proto_imf, proto_imf_prev, rtol=0.01):
            return True

        # If IMF mean close to zero (below threshold)
        if np.mean(np.abs(proto_imf)) < self.mean_thr:
            return True

        # Everything relatively close to 0
        mse_proto_imf = np.mean(proto_imf * proto_imf)
        if mse_proto_imf > self.mse_thr:
            return False

        return False

    def bemd(self, image, max_imf=-1):
        """Performs bidimensional EMD (BEMD) on grey-scale image with specified parameters.

        Parameters
        ----------
        image : numpy 2D array,
            Grey-scale image.
        max_imf : int, (default: -1)
            IMF number to which decomposition should be performed.
            Negative value means *all*.

        Returns
        -------
        IMFs : numpy 3D array
            Set of IMFs in form of numpy array where the first dimension
            relates to IMF's ordinary number.
        """
        image_s = image.copy()

        imf = np.zeros(image.shape)
        imf_old = imf.copy()

        imfNo = 0
        IMF = np.empty((imfNo,) + image.shape)
        notFinished = True

        while notFinished:
            self.logger.debug("IMF -- " + str(imfNo))

            res = image_s - np.sum(IMF[:imfNo], axis=0)
            imf = res.copy()
            mean_env = np.zeros(image.shape)
            stop_sifting = False

            # Counters
            n = 0  # All iterations for current imf.
            n_h = 0  # counts when mean(proto_imf) < threshold

            while not stop_sifting and n < self.MAX_ITERATION:
                n += 1
                self.logger.debug("Iteration: %i", n)

                min_peaks_pos, max_peaks_pos = self.find_extrema_positions(imf)
                self.logger.debug(
                    "min_peaks_pos = %i  |  max_peaks_pos = %i", len(min_peaks_pos[0]), len(max_peaks_pos[0])
                )
                if len(min_peaks_pos[0]) > 1 and len(max_peaks_pos[0]) > 1:
                    min_env, max_env = self.extract_max_min_spline(imf, min_peaks_pos, max_peaks_pos)
                    mean_env = 0.5 * (min_env + max_env)

                    imf_old = imf.copy()
                    imf = imf - mean_env

                    # Fix number of iterations
                    if self.FIXE:
                        if n >= self.FIXE + 1:
                            stop_sifting = True

                    # Fix number of iterations after number of zero-crossings
                    # and extrema differ at most by one.
                    elif self.FIXE_H:
                        if n == 1:
                            continue
                        if self.check_proto_imf(imf, imf_old, mean_env):
                            n_h += 1
                        else:
                            n_h = 0

                        # STOP if enough n_h
                        if n_h >= self.FIXE_H:
                            stop_sifting = True

                    # Stops after default stopping criteria are met
                    else:
                        if self.check_proto_imf(imf, imf_old, mean_env):
                            stop_sifting = True

                else:
                    stop_sifting = True

            IMF = np.vstack((IMF, imf.copy()[None, :]))
            imfNo += 1

            if self.end_condition(image, IMF) or (max_imf > 0 and imfNo >= max_imf):
                notFinished = False
                break

        res = image_s - np.sum(IMF[:imfNo], axis=0)
        if not np.allclose(res, 0):
            IMF = np.vstack((IMF, res[None, :]))
            imfNo += 1

        return IMF


if __name__ == "__main__":
    print("Running example on BEMD")
    PLOT = True

    logging.basicConfig(level=logging.DEBUG)

    # Generate image
    print("Generating image... ", end="")
    rows, cols = 256, 256
    row_scale, col_scale = 256, 256
    x = np.arange(rows) / float(row_scale)
    y = np.arange(cols).reshape((-1, 1)) / float(col_scale)

    pi2 = 2 * np.pi
    img = np.zeros((rows, cols))
    # img = img + np.sin(2*pi2*x)*np.cos(y*4*pi2+4*x*pi2)
    img = img + 3 * np.sin(2 * pi2 * x) + 2
    # img = img + 5*x*y + 2*(y-0.2)*y
    print("Done")

    # Perform decomposition
    print("Performing decomposition... ", end="")
    bemd = BEMD()
    # bemd.FIXE_H = 5
    IMFs = bemd.bemd(img, max_imf=3)
    imfNo = IMFs.shape[0]
    print("Done")

    if PLOT:
        print("Plotting results... ", end="")
        import pylab as plt

        # Save image for preview
        plt.figure(figsize=(4, 4 * (imfNo + 1)))
        plt.subplot(imfNo + 1, 1, 1)
        plt.imshow(img)
        plt.colorbar()
        plt.title("Input image")

        # Save reconstruction
        for n, imf in enumerate(IMFs):
            plt.subplot(imfNo + 1, 1, n + 2)
            plt.imshow(imf)
            plt.colorbar()
            plt.title("IMF %i" % (n + 1))

        plt.savefig("image_decomp")
        print("Done")



# --------------------------------------------------------------------------------------------------------
# WaLSAtools - Wave analysis tools
# Copyright (C) 2025 WaLSA Team - Shahin Jafarzadeh et al.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Note: If you use WaLSAtools for research, please consider citing:
# Jafarzadeh, S., Jess, D. B., Stangalini, M. et al. 2025, Nature Reviews Methods Primers, in press.
# --------------------------------------------------------------------------------------------------------
# The following codes are based on those originally written by Jonathan E. Higham & Luiz A. C. A. Schiavo
# --------------------------------------------------------------------------------------------------------

import numpy as np # type: ignore
from scipy.signal import welch, find_peaks # type: ignore
from scipy.linalg import svd # type: ignore
from scipy.optimize import curve_fit # type: ignore
import math

def print_pod_results(results):
    """
    Print a summary of the results from WaLSA_pod, including parameter descriptions, types, and shapes.
    
    Parameters:
    -----------
    results : dict
        The dictionary containing all POD results and relevant outputs.
    """
    descriptions = {
        'input_data': 'Original input data, mean subtracted (Shape: (Nt, Ny, Nx))',
        'spatial_mode': 'Reshaped spatial modes matching the dimensions of the input data (Shape: (Nmodes, Ny, Nx))',
        'temporal_coefficient': 'Temporal coefficients associated with each spatial mode (Shape: (Nmodes, Nt))',
        'eigenvalue': 'Eigenvalues corresponding to singular values squared (Shape: (Nmodes))',
        'eigenvalue_contribution': 'Eigenvalue contribution of each mode (Shape: (Nmodes))',
        'cumulative_eigenvalues': 'Cumulative percentage of eigenvalues for the first "num_cumulative_modes" modes (Shape: (num_cumulative_modes))',
        'combined_welch_psd': 'Combined Welch power spectral density for the temporal coefficients of the firts "num_modes" modes (Shape: (Nf))',
        'frequencies': 'Frequencies identified in the Welch spectrum (Shape: (Nf))',
        'combined_welch_significance': 'Significance threshold of the combined Welch spectrum (Shape: (Nf,))',
        'reconstructed': 'Reconstructed frame at the specified timestep (or for the entire time series) using the top "num_modes" modes (Shape: (Ny, Nx))',
        'sorted_frequencies': 'Frequencies identified in the Welch combined power spectrum (Shape: (Nfrequencies))',
        'frequency_filtered_modes': 'Frequency-filtered spatial POD modes for the first "num_top_frequencies" frequencies (Shape: (Nt, Ny, Nx, num_top_frequencies))',
        'frequency_filtered_modes_frequencies': 'Frequencies corresponding to the frequency-filtered modes (Shape: (num_top_frequencies))',
        'SPOD_spatial_modes': 'SPOD spatial modes if SPOD is used (Shape: (Nspod_modes, Ny, Nx))',
        'SPOD_temporal_coefficients': 'SPOD temporal coefficients if SPOD is used (Shape: (Nspod_modes, Nt))',
        'p': 'Left singular vectors (spatial modes) from SVD (Shape: (Nx, Nmodes))',
        's': 'Singular values from SVD (Shape: (Nmodes))',
        'a': 'Right singular vectors (temporal coefficients) from SVD (Shape: (Nmodes, Nt))'
    }
    
    print("\n---- POD/SPOD Results Summary ----\n")
    for key, value in results.items():
        desc = descriptions.get(key, 'No description available')
        shape = np.shape(value) if value is not None else 'None'
        dtype = type(value).__name__
        print(f"{key} ({dtype}, Shape: {shape}): {desc}")
    print("\n----------------------------------")

def spod(data, **kwargs):
    """
    Perform Spectral Proper Orthogonal Decomposition (SPOD) analysis on input data.
    Steps:
    1. Load the data.
    2. Compute a time average.
    3. Build the correlation matrix to compute the POD using the snapshot method.
    4. Compute eigenvalue decomposition for the correlation matrix for the fluctuation field. 
    The eigenvalues and eigenvectors can be computed by any eigenvalue function or by an SVD procedure, since the correlation matrix is symmetric positive-semidefinite.
    5. After obtaining the eigenvalues and eigenvectors, compute the temporal and spatial modes according to the snapshot method.
    
    Parameters:
    -----------
    data : np.ndarray
        3D data array with shape (time, x, y) or similar. 
    **kwargs : dict, optional
        Additional keyword arguments to configure the analysis.
    
    Returns:
    --------
    SPOD_spatial_modes : np.ndarray
    SPOD_temporal_coefficients : np.ndarray
    """
    # Set default parameter values
    defaults = {
        'silent': False,
        'num_modes': None,
        'filter_size': None,
        'periodic_matrix': True
    }
    
    # Update defaults with user-provided values
    params = {**defaults, **kwargs}

    if params['num_modes'] is None:
        params['num_modes'] = data.shape[0]
    if params['filter_size'] is None:
        params['filter_size'] = params['num_modes']

    if not params['silent']:
        print("Starting SPOD analysis ....")

    nsnapshots, ny, nx = data.shape

    # Center the data by subtracting the mean
    time_average = data.mean(axis=0)
    fluctuation_field = data * 0  # Initialize fluctuation field with zeros
    for n in range(nsnapshots):
        fluctuation_field[n,:,:] = data[n,:,:] - time_average  # Subtract time-average from each snapshot

    # Build the correlation matrix (snapshot method)
    correlation_matrix = np.zeros((nsnapshots, nsnapshots))
    for i in range(nsnapshots):
        for j in range(i, nsnapshots):
            correlation_matrix[i, j] = (fluctuation_field[i, :, :] * fluctuation_field[j, :, :]).sum() / (nsnapshots * nx * ny)
            correlation_matrix[j, i] = correlation_matrix[i, j]

    # SPOD correlation matrix with periodic boundary conditions
    nmatrix = np.zeros((3 * nsnapshots, 3 * nsnapshots))
    nmatrix[nsnapshots:2 * nsnapshots, nsnapshots:2 * nsnapshots] = correlation_matrix[0:nsnapshots, 0:nsnapshots]

    if params['periodic_matrix']:
        for i in range(3):
            for j in range(3):
                xs = i * nsnapshots  # x-offset for periodic positioning
                ys = j * nsnapshots  # y-offset for periodic positioning
                nmatrix[0 + xs:nsnapshots + xs, 0 + ys:nsnapshots + ys] = correlation_matrix[0:nsnapshots, 0:nsnapshots]

    # Apply Gaussian filter
    gk = np.zeros((2 * params['filter_size'] + 1))  # Create 1D Gaussian kernel
    esp = 8.0  # Exponential parameter controlling the spread of the filter
    sumgk = 0.0  # Sum of the Gaussian kernel for normalization

    for n in range(2 * params['filter_size'] + 1):
        k = -params['filter_size'] + n  # Offset from the center of the kernel
        gk[n] = math.exp(-esp * k**2.0 / (params['filter_size']**2.0))  # Gaussian filter formula
        sumgk += gk[n]  # Sum of kernel values for normalization

    # Filter the extended correlation matrix
    for j in range(nsnapshots, 2 * nsnapshots):
        for i in range(nsnapshots, 2 * nsnapshots):
            aux = 0.0  # Initialize variable for the weighted sum
            for n in range(2 * params['filter_size'] + 1):
                k = -params['filter_size'] + n  # Offset from the current index
                aux += nmatrix[i + k, j + k] * gk[n]  # Apply Gaussian weighting
            nmatrix[i, j] = aux / sumgk  # Normalize by the sum of the Gaussian kernel
    
    # Extract the SPOD correlation matrix from the central part of the filtered matrix
    spectral_matrix = nmatrix[nsnapshots:2 * nsnapshots, nsnapshots:2 * nsnapshots]

    # Perform SVD to compute SPOD
    UU, SS, VV = svd(spectral_matrix, full_matrices=True, compute_uv=True)

    # Temporal coefficients
    SPOD_temporal_coefficients = np.zeros((nsnapshots, nsnapshots))
    for k in range(nsnapshots):
        SPOD_temporal_coefficients[:, k] = np.sqrt(SS[k] * nsnapshots) * UU[:, k]

    # Extract spatial SPOD modes
    SPOD_spatial_modes = np.zeros((params['num_modes'], ny, nx))
    for m in range(params['num_modes']):
        for t in range(nsnapshots):
            SPOD_spatial_modes[m, :, :] += fluctuation_field[t, :, :] * SPOD_temporal_coefficients[t, m] / (SS[m] * nsnapshots)

    if not params['silent']:
        print(f"SPOD analysis completed.")
    
    return SPOD_spatial_modes, SPOD_temporal_coefficients


# -------------------------------------- Main Function ----------------------------------------
def WaLSA_pod(signal, time, **kwargs):
    """
    Perform Proper Orthogonal Decomposition (POD) analysis on input data.
    
    Parameters:
        signal (array): 3D data cube with shape (time, x, y) or similar.
        time (array): 1D array representing the time points for each time step in the data.
        num_modes (int, optional): Number of top modes to compute. Default is None (all modes).
        num_top_frequencies (int, optional): Number of top frequencies to consider. Default is None (all frequencies).
        top_frequencies (list, optional): List of top frequencies to consider. Default is None.
        num_cumulative_modes (int, optional): Number of cumulative modes to consider. Default is None (all modes).
        welch_nperseg (int, optional): Number of samples per segment for Welch's method. Default is 150.
        welch_noverlap (int, optional): Number of overlapping samples for Welch's method. Default is 25.
        welch_nfft (int, optional): Number of points for the FFT. Default is 2^14.
        welch_fs (int, optional): Sampling frequency for the data. Default is 2.
        nperm (int, optional): Number of permutations for significance testing. Default is 1000.
        siglevel (float, optional): Significance level for the Welch spectrum. Default is 0.95.
        timestep_to_reconstruct (int, optional): Timestep of the datacube to reconstruct using the top modes. Default is 0.
        num_modes_reconstruct (int, optional): Number of modes to use for reconstruction. Default is None (all modes).
        reconstruct_all (bool, optional): If True, reconstruct the entire time series using the top modes. Default is False.
        spod (bool, optional): If True, perform Spectral Proper Orthogonal Decomposition (SPOD) analysis. Default is False.
        spod_filter_size (int, optional): Filter size for SPOD analysis. Default is None.
        spod_num_modes (int, optional): Number of SPOD modes to compute. Default is None.
        print_results (bool, optional): If True, print a summary of results. Default is True.

    **kwargs : Additional keyword arguments to configure the analysis.
    
    Returns:
    results : dict
        A dictionary containing all computed POD results and relevant outputs. 
        See 'descriptions' in the 'print_pod_results' function on top of this page.
    """
    # Set default parameter values
    defaults = {
        'silent': False,
        'num_modes': None,
        'num_top_frequencies': None,
        'top_frequencies': None,
        'num_cumulative_modes': None,
        'welch_nperseg': 150,
        'welch_noverlap': 25,
        'welch_nfft': 2**14,
        'welch_fs': 2,
        'nperm': 1000,
        'siglevel': 0.95,
        'timestep_to_reconstruct': 0,
        'reconstruct_all': False,
        'num_modes_reconstruct': None,
        'spod': False,
        'spod_filter_size': None,
        'spod_num_modes': None,
        'print_results': True  # Print summary of results by default
    }
    
    # Update defaults with user-provided values
    params = {**defaults, **kwargs}

    data = signal

    if params['num_modes'] is None:
        params['num_modes'] = data.shape[0]

    if params['num_top_frequencies'] is None:
        params['num_top_frequencies'] = min(params['num_modes'], 10)
    
    if params['num_cumulative_modes'] is None:
        params['num_cumulative_modes'] = min(params['num_modes'], 10)

    if params['num_modes_reconstruct'] is None:
        params['num_modes_reconstruct'] = min(params['num_modes'], 10)

    if not params['silent']:
        print("Starting POD analysis ....")
        print(f"Processing a 3D cube with shape {data.shape}.")

    # The first step is to read in the data and then perform the SVD (or POD). Before we do this, 
    # we need to reshape the matrix such that it is an N x T matrix where N is the column vectorized set of each spatial image and T is the temporal domain. 
    # We also need to ensure that the mean is subtracted from the data; this will ensure we are looking only at the variance of the data, and mode 1 will not be contaminated with the mean image.
    # Reshape the 3D data into 2D array where each row is a vector from the original 3D data
    inp = np.reshape(data, (data.shape[0], data.shape[1] * data.shape[2])).astype(np.float32)
    inp = inp.T  # Transpose the matrix to have spatial vectors as columns

    # Center the data by subtracting the mean
    mean_per_row = np.nanmean(inp, axis=1, keepdims=True)
    mean_replicated = np.tile(mean_per_row, (1, data.shape[0]))
    inp_centered = inp - mean_replicated

    # Input data, mean subtracted
    input_dat = np.zeros((data.shape[0], data.shape[1], data.shape[2]))
    for im in range(data.shape[0]):
        input_dat[im, :, :] = np.reshape(inp_centered[:, im], (data.shape[1], data.shape[2]))

    # Perform SVD to compute POD
    p, s, a = svd(inp_centered, full_matrices=False)
    sorg = s.copy()  # Store original singular values
    eigenvalue = s**2  # Convert singular values to eigenvalues

    # Reshape spatial modes to have the same shape as the input data
    num_modes = p.shape[1]
    spatial_mode = np.zeros((num_modes, data.shape[1], data.shape[2]))
    for m in range(num_modes):
        spatial_mode[m, :, :] = np.reshape(p[:, m], (data.shape[1], data.shape[2]))

    # Extract temporal coefficients
    temporal_coefficient = a[:num_modes, :]

    # Calculate eigenvalue contributions
    eigenvalue_contribution = eigenvalue / np.sum(eigenvalue)

    # Calculate cumulative eigenvalues
    cumulative_eigenvalues = []
    for m in range(params['num_cumulative_modes']): 
        contm = 100 * eigenvalue[0:m] / np.sum(eigenvalue)
        cumulative_eigenvalues.append(np.sum(contm)) 

    if params['reconstruct_all']:
        # Reconstruct the entire time series using the top 'num_modes_reconstruct' modes
        reconstructed = np.zeros((data.shape[0], data.shape[1], data.shape[2]))
        for tindex in range(data.shape[0]):
            reconim = np.zeros((data.shape[1], data.shape[2]))
            for i in range(params['num_modes_reconstruct']):
                reconim=reconim+np.reshape(p[:, i], (data.shape[1], data.shape[2]))*a[i, tindex]*sorg[i]
            reconstructed[tindex, :, :] = reconim
    else:
        # Reconstruct the specified timestep using the top 'num_modes_reconstruct' modes
        reconstructed = np.zeros((data.shape[1], data.shape[2]))
        for i in range(params['num_modes_reconstruct']):
            reconstructed=reconstructed+np.reshape(p[:, i], (data.shape[1], data.shape[2]))*a[i, params['timestep_to_reconstruct']]*sorg[i]

    #---------------------------------------------------------------------------------
    # Combined Welch power spectrum and its significance
    #---------------------------------------------------------------------------------
    # Compute Welch power spectrum for each mode and combine them for 'num_modes' modes
    combined_welch_psd = []
    for m in range(params['num_modes']):
        frequencies, px = welch(a[m, :] - np.mean(a[m, :]), 
                      nperseg=params['welch_nperseg'], 
                      noverlap=params['welch_noverlap'], 
                      nfft=params['welch_nfft'], 
                      fs=params['welch_fs'])
        if m == 0:
            combined_welch_psd = np.zeros((len(frequencies),))
        combined_welch_psd += eigenvalue_contribution[m] * (px / np.sum(px))

    # Generate resampled peaks to compute significance threshold
    resampled_peaks = np.zeros((params['nperm'], len(frequencies)))
    for i in range(params['nperm']):
        resampled_data = np.random.randn(*a.shape)
        resampled_peak = np.zeros((len(frequencies),))
        for m in range(params['num_modes']):
            _, px = welch(resampled_data[m, :] - np.mean(resampled_data[m, :]), 
                          nperseg=params['welch_nperseg'], 
                          noverlap=params['welch_noverlap'], 
                          nfft=params['welch_nfft'], 
                          fs=params['welch_fs'])
            resampled_peak += eigenvalue_contribution[m] * (px / np.sum(px))
        resampled_peak /= (np.max(resampled_peak) + 1e-30) # a small epsilon added to avoid division by zero
        resampled_peaks[i, :] = resampled_peak
    # Calculate significance threshold
    combined_welch_significance = np.percentile(resampled_peaks, 100-(params['siglevel']*100), axis=0)
    #---------------------------------------------------------------------------------

    # Find peaks in the combined spectrum and sort them in descending order
    normalized_peak = combined_welch_psd / (np.max(combined_welch_psd) + 1e-30)
    peak_indices, _ = find_peaks(normalized_peak)
    sorted_indices = np.argsort(normalized_peak[peak_indices])[::-1]
    sorted_frequencies = frequencies[peak_indices][sorted_indices]

    # Generate a table of the top N frequencies
    # Cleaner single-line list comprehension
    table_data = [
        [float(np.round(freq, 2)), float(np.round(pwr, 2))] 
        for freq, pwr in zip(sorted_frequencies[:params['num_top_frequencies']], 
                            normalized_peak[peak_indices][sorted_indices][:params['num_top_frequencies']])
    ]

    # The frequencies that the POD is able to capture have now been identified (in the top 'num_top_frequencies' modes). 
    # These frequencies can be fitted to the temporal coefficients of the top 'num_top_frequencies' modes, 
    # allowing for a representation of the data described solely by these "pure" frequencies. 
    # This approach enables the reconstruction of the original data using the identified dominant frequencies,
    # resulting in frequency-filtered spatial POD modes.
    clean_data = np.zeros((inp.shape[0],inp.shape[1],10))

    if params['top_frequencies'] is None:
        top_frequencies = sorted_frequencies[:params['num_top_frequencies']]
    else:    
        top_frequencies = params['top_frequencies']
        params['num_top_frequencies'] = len(params['top_frequencies'])

    for i in range(params['num_top_frequencies']):
        def model_fun(t, amplitude, phase):
            """
            Generate a sinusoidal model function.

            Parameters:
            t (array-like): The time variable.
            amplitude (float): The amplitude of the sinusoidal function.
            phase (float): The phase shift of the sinusoidal function.

            Returns:
            array-like: The computed sinusoidal values at each time point `t`.
            """
            return amplitude * np.sin(2 * np.pi * top_frequencies[i] * t + phase)
        
        for j in range(params['num_modes_reconstruct']):
            csignal = a[j,:]
            # Initial Guesses for Parameters: [Amplitude, Phase]
            initial_guess = [1, 0]
            # Nonlinear Fit
            fit_params, _ = curve_fit(model_fun, time, csignal, p0=initial_guess)
            # Clean Signal from Fit
            if j == 0:
                clean_signal=np.zeros((params['num_modes_reconstruct'],len(csignal)))
            clean_signal[j,:] = model_fun(time, *fit_params)
        # forming a set of clean data (reconstructed data at the fitted frequencies; frequency filtered reconstructed data)
        clean_data[:,:,i] = p[:,0:params['num_modes_reconstruct']]@np.diag(sorg[0:params['num_modes_reconstruct']])@clean_signal

    frequency_filtered_modes = np.zeros((data.shape[0], data.shape[1], data.shape[2], params['num_top_frequencies']))
    for jj in range(params['num_top_frequencies']):
        for frame_index in range(data.shape[0]):
            frequency_filtered_modes[frame_index, :, :, jj] = np.reshape(clean_data[:, frame_index, jj], (data.shape[1], data.shape[2]))

    if params['spod']:
        SPOD_spatial_modes, SPOD_temporal_coefficients = spod(data, num_modes=params['spod_num_modes'], filter_size=params['spod_filter_size'])
    else:
        SPOD_spatial_modes = None
        SPOD_temporal_coefficients = None

    results = {
        'input_data': input_dat, # Original input data, mean subtracted (Shape: (Nt, Ny, Nx))
        'spatial_mode': spatial_mode,  # POD spatial modes matching the dimensions of the input data (Shape: (Nmodes, Ny, Nx))
        'temporal_coefficient': temporal_coefficient,  # POD temporal coefficients associated with each spatial mode (Shape: (Nmodes, Nt))
        'eigenvalue': eigenvalue,  # Eigenvalues corresponding to singular values squared (Shape: (Nmodes))
        'eigenvalue_contribution': eigenvalue_contribution,  # Eigenvalue contribution of each mode (Shape: (Nmodes))
        'cumulative_eigenvalues': cumulative_eigenvalues,  # Cumulative percentage of eigenvalues for the first 'num_cumulative_modes' modes (Shape: (Ncumulative_modes))
        'combined_welch_psd': combined_welch_psd,  # Combined Welch power spectral density for the temporal coefficients (Shape: (Nf))
        'frequencies': frequencies,  # Frequencies identified in the Welch spectrum (Shape: (Nf))
        'combined_welch_significance': combined_welch_significance,  # Significance threshold of the combined Welch spectrum (Shape: (Nf))
        'reconstructed': reconstructed,  # Reconstructed frame at the specified timestep (or for the entire time series) using the top modes (Shape: (Ny, Nx))
        'sorted_frequencies': sorted_frequencies,  # Frequencies identified in the Welch spectrum (Shape: (Nfrequencies,))
        'frequency_filtered_modes': frequency_filtered_modes,  # Frequency-filtered spatial POD modes (Shape: (Nt, Ny, Nx, Ntop_frequencies))
        'frequency_filtered_modes_frequencies': top_frequencies,  # Frequencies corresponding to the frequency-filtered modes (Shape: (Ntop_frequencies))
        'SPOD_spatial_modes': SPOD_spatial_modes,  # SPOD spatial modes if SPOD is used (Shape: (Nspod_modes, Ny, Nx))
        'SPOD_temporal_coefficients': SPOD_temporal_coefficients,  # SPOD temporal coefficients if SPOD is used (Shape: (Nspod_modes, Nt))
        'p': p,  # Left singular vectors (spatial modes) from SVD (Shape: (Nx, Nmodes))
        's': sorg,  # Singular values from SVD (Shape: (Nmodes,))
        'a': a  # Right singular vectors (temporal coefficients) from SVD (Shape: (Nmodes, Nt))
    }

    if not params['silent']:
        print(f"POD analysis completed.")
        print(f"Top {params['num_top_frequencies']} frequencies and normalized power values:\n{table_data}")
        print(f"Total variance contribution of the first {params['num_modes']} modes: {np.sum(100 * eigenvalue[:params['num_modes']] / np.sum(eigenvalue)):.2f}%")

    if params['print_results']:
        print_pod_results(results)

    return results



# -----------------------------------------------------------------------------------------------------
# WaLSAtools - Wave analysis tools
# Copyright (C) 2025 WaLSA Team - Shahin Jafarzadeh et al.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Note: If you use WaLSAtools for research, please consider citing:
# Jafarzadeh, S., Jess, D. B., Stangalini, M. et al. 2025, Nature Reviews Methods Primers, in press.
# -----------------------------------------------------------------------------------------------------

import numpy as np # type: ignore
from .analysis_modules.WaLSA_speclizer import WaLSA_speclizer
from .analysis_modules.WaLSA_k_omega import WaLSA_k_omega
from .analysis_modules.WaLSA_pod import WaLSA_pod
from .analysis_modules.WaLSA_cross_spectra import WaLSA_cross_spectra
from .analysis_modules.WaLSA_interactive import interactive

class WaLSAtools:
    """
    Main class to perform spectral analysis using different methods.
    
    It runs interactively if called without (or empty
    and acts as a function if called with parentheses and parameters.
    """
    
    def __call__(self, signal=None, time=None, method=None, **kwargs):
        """
        Main function to perform spectral analysis using different methods.
        
        Parameters:
            signal (array): The input signal (1D, or 3D).
            time (array): The time array of the signal.
            method (str): The method to use for spectral analysis ('fft', 'lombscargle', 'k-omega', etc.)
            **kwargs: Additional parameters for data preparation and analysis methods.
        
        Returns:
            Results of the selected analysis method.
        """
        if signal is None and time is None and method is None:
            interactive()  # Run interactive (help) mode if no inputs are provided
            return None
        
        method_map = {
            'fft': WaLSA_speclizer,
            'lombscargle': WaLSA_speclizer,
            'wavelet': WaLSA_speclizer,
            'welch': WaLSA_speclizer,
            'emd': WaLSA_speclizer,
            'k-omega': WaLSA_k_omega,
            'pod': WaLSA_pod
        }
        
        if method not in method_map:
            raise ValueError(f"Unknown method '{method}'. Please choose from {list(method_map.keys())}.")
        
        func = method_map[method]
        
        if method in ['fft', 'lombscargle', 'wavelet', 'welch', 'emd'] and 'data1' in kwargs and 'data2' in kwargs:
            return WaLSA_cross_spectra(signal=np.ones(10), time=time, method=method, **kwargs)
        else:
            return func(signal=signal, time=time, method=method, **kwargs)
    
    def __repr__(self):
        """
        Custom representation to allow interactive mode when 'WaLSAtools' is typed without parentheses.
        """
        interactive()  # Call the interactive (help) function
        return ''  # Return an empty string so that no additional output is shown

# Instantiate the object so that it works like a function and runs interactively
WaLSAtools = WaLSAtools()


# -----------------------------------------------------------------------------------------------------
# WaLSAtools - Wave analysis tools
# Copyright (C) 2025 WaLSA Team - Shahin Jafarzadeh et al.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Note: If you use WaLSAtools for research, please consider citing:
# Jafarzadeh, S., Jess, D. B., Stangalini, M. et al. 2025, Nature Reviews Methods Primers, in press.
# -----------------------------------------------------------------------------------------------------

from setuptools import setup, find_packages

# Read dependencies from requirements.txt
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

# Read long description safely
try:
    with open('README.md', encoding='utf-8') as f:
        long_description = f.read()
except FileNotFoundError:
    long_description = "Wave analysis tools: A package for analyzing wave and oscillatory signals."

setup(
    name='WaLSAtools',  # Package name on PyPI
    version='1.0.0.post1',  # Update version as needed
    author='WaLSA Team - Shahin Jafarzadeh',
    author_email='Shahin.Jafarzadeh@WaLSA.team',
    description='Wave analysis tools: A package for analyzing wave and oscillatory signals',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/WaLSAteam/WaLSAtools',
    packages=find_packages(),
    include_package_data=True, 
    package_data={
        "WaLSAtools": ["assets/*.png", "assets/*.jpg"],
    },
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            'WaLSAtools=WaLSAtools.WaLSAtools:main', 
        ],
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',  
) 



import numpy as np
from scipy.interpolate import Akima1DInterpolator, CubicHermiteSpline, CubicSpline, PchipInterpolator


def cubic_spline_3pts(x, y, T):
    """
    Apparently scipy.interpolate.interp1d does not support
    cubic spline for less than 4 points.
    """
    x0, x1, x2 = x
    y0, y1, y2 = y

    x1x0, x2x1 = x1 - x0, x2 - x1
    y1y0, y2y1 = y1 - y0, y2 - y1
    _x1x0, _x2x1 = 1.0 / x1x0, 1.0 / x2x1

    m11, m12, m13 = 2 * _x1x0, _x1x0, 0
    m21, m22, m23 = _x1x0, 2.0 * (_x1x0 + _x2x1), _x2x1
    m31, m32, m33 = 0, _x2x1, 2.0 * _x2x1

    v1 = 3 * y1y0 * _x1x0 * _x1x0
    v3 = 3 * y2y1 * _x2x1 * _x2x1
    v2 = v1 + v3

    M = np.array([[m11, m12, m13], [m21, m22, m23], [m31, m32, m33]])
    v = np.array([v1, v2, v3]).T
    k = np.linalg.inv(M).dot(v)

    a1 = k[0] * x1x0 - y1y0
    b1 = -k[1] * x1x0 + y1y0
    a2 = k[1] * x2x1 - y2y1
    b2 = -k[2] * x2x1 + y2y1

    t = T[np.r_[T >= x0] & np.r_[T <= x2]]
    t1 = (T[np.r_[T >= x0] & np.r_[T < x1]] - x0) / x1x0
    t2 = (T[np.r_[T >= x1] & np.r_[T <= x2]] - x1) / x2x1
    t11, t22 = 1.0 - t1, 1.0 - t2

    q1 = t11 * y0 + t1 * y1 + t1 * t11 * (a1 * t11 + b1 * t1)
    q2 = t22 * y1 + t2 * y2 + t2 * t22 * (a2 * t22 + b2 * t2)
    q = np.append(q1, q2)

    return t, q


def akima(X, Y, x):
    spl = Akima1DInterpolator(X, Y)
    return spl(x)


def cubic_hermite(X, Y, x):
    dydx = np.gradient(Y, X)
    spl = CubicHermiteSpline(X, Y, dydx)
    return spl(x)


def cubic(X, Y, x):
    spl = CubicSpline(X, Y)
    return spl(x)


def pchip(X, Y, x):
    spl = PchipInterpolator(X, Y)
    return spl(x)



import logging

__version__ = "1.6.4"
logger = logging.getLogger("pyemd")

from .CEEMDAN import CEEMDAN  # noqa
from .EEMD import EEMD  # noqa
from .EMD import EMD  # noqa
from .visualisation import Visualisation  # noqa



# -----------------------------------------------------------------------------------------------------
# WaLSAtools - Wave analysis tools
# Copyright (C) 2025 WaLSA Team - Shahin Jafarzadeh et al.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Note: If you use WaLSAtools for research, please consider citing:
# Jafarzadeh, S., Jess, D. B., Stangalini, M. et al. 2025, Nature Reviews Methods Primers, in press.
# -----------------------------------------------------------------------------------------------------

import ipywidgets as widgets # type: ignore
from IPython.display import display, clear_output, HTML # type: ignore
import os # type: ignore

# Function to detect if it's running in a notebook or terminal
def is_notebook():
    try:
        # Check if IPython is in the environment and if we're using a Jupyter notebook
        from IPython import get_ipython # type: ignore
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True  # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # IPython terminal
        else:
            return False  # Other types
    except (NameError, ImportError):
        # NameError: get_ipython is not defined
        # ImportError: IPython is not installed
        return False  # Standard Python interpreter or shell

# Function to print logo and credits for both environments
def print_logo_and_credits():
    logo_terminal = r"""
        __          __          _          _____            
        \ \        / /         | |        / ____|     /\    
         \ \  /\  / /  ▄▄▄▄▄   | |       | (___      /  \   
          \ \/  \/ /   ▀▀▀▀██  | |        \___ \    / /\ \  
           \  /\  /   ▄██▀▀██  | |____    ____) |  / ____ \ 
            \/  \/    ▀██▄▄██  |______|  |_____/  /_/    \_\
    """
    credits_terminal = """
        © WaLSA Team (www.WaLSA.team)
        -----------------------------------------------------------------------
        WaLSAtools v1.0 - Wave analysis tools
        Documentation: www.WaLSA.tools
        GitHub repository: www.github.com/WaLSAteam/WaLSAtools
        -----------------------------------------------------------------------
        If you use WaLSAtools in your research, please cite:
        Jafarzadeh, S., Jess, D. B., Stangalini, M. et al. 2025,
        Nature Reviews Methods Primers, in press
        -----------------------------------------------------------------------
        Choose a category, data type, and analysis method from the list below,
        to get hints on the calling sequence and parameters:
        """

    credits_notebook = """
        <div style="margin-left: 30px; margin-top: 20px; font-size: 1.1em; line-height: 0.8;">
            <p>© WaLSA Team (<a href="https://www.WaLSA.team" target="_blank">www.WaLSA.team</a>)</p>
            <hr style="width: 70%; margin: 0; border: 0.98px solid #888; margin-bottom: 10px;">
            <p><strong>WaLSAtools</strong> v1.0 - Wave analysis tools</p>
            <p>Documentation: <a href="https://www.WaLSA.tools" target="_blank">www.WaLSA.tools</a></p>
            <p>GitHub repository: <a href="https://www.github.com/WaLSAteam/WaLSAtools" target="_blank">www.github.com/WaLSAteam/WaLSAtools</a></p>
            <hr style="width: 70%; margin: 0; border: 0.98px solid #888; margin-bottom: 10px;">
            <p>If you use <strong>WaLSAtools</strong> in your research, please cite:</p>
            <p>Jafarzadeh, S., Jess, D. B., Stangalini, M. et al. 2025, <em>Nature Reviews Methods Primers</em>, in press</p>
            <hr style="width: 70%; margin: 0; border: 0.98px solid #888; margin-bottom: 15px;">
            <p>Choose a category, data type, and analysis method from the list below,</p>
            <p>to get hints on the calling sequence and parameters:</p>
        </div>
    """

    if is_notebook():
        try:
            # For scripts
            current_dir = os.path.dirname(os.path.abspath(__file__))
        except NameError:
            # For Jupyter notebooks
            current_dir = os.getcwd()
        img_path = os.path.join(current_dir, '..', 'assets', 'WaLSAtools_black.png')
        # display(HTML(f'<img src="{img_path}" style="margin-left: 40px; margin-top: 20px; width:300px; height: auto;">')) # not shwon in Jupyter notebook, only in MS Code
        import base64
        # Convert the image to Base64
        with open(img_path, "rb") as img_file:
            encoded_img = base64.b64encode(img_file.read()).decode('utf-8')
        # Embed the Base64 image in the HTML
        html_code_logo = f"""
        <div style="margin-left: 30px; margin-top: 20px;">
            <img src="data:image/png;base64,{encoded_img}" style="width: 300px; height: auto;">
        </div>
        """
        display(HTML(html_code_logo))
        display(HTML(credits_notebook))
    else:
        print(logo_terminal)
        print(credits_terminal)

from .parameter_definitions import display_parameters_text, single_series_parameters, cross_correlation_parameters

# Terminal-based interactive function
def walsatools_terminal():
    """Main interactive function for terminal version of WaLSAtools."""
    print_logo_and_credits()

    # Step 1: Select Category
    while True:
        print("\n    Category:")
        print("    (a) Single time series analysis")
        print("    (b) Cross-correlation between two time series")
        category = input("    --- Select a category (a/b): ").strip().lower()

        if category not in ['a', 'b']:
            print("    Invalid selection. Please enter either 'a' or 'b'.\n")
            continue

        # Step 2: Data Type
        if category == 'a':
            while True:
                print("\n    Data Type:")
                print("    (1) 1D signal")
                print("    (2) 3D datacube")
                method = input("    --- Select a data type (1/2): ").strip()

                if method not in ['1', '2']:
                    print("    Invalid selection. Please enter either '1' or '2'.\n")
                    continue

                # Step 3: Analysis Method
                if method == '1':  # 1D Signal
                    while True:
                        print("\n    Analysis Method:")
                        print("    (1) FFT")
                        print("    (2) Wavelet")
                        print("    (3) Lomb-Scargle")
                        print("    (4) Welch")
                        print("    (5) EMD")
                        analysis_type = input("    --- Select an analysis method (1-5): ").strip()

                        method_map_1d = {
                            '1': 'fft',
                            '2': 'wavelet',
                            '3': 'lombscargle',
                            '4': 'welch',
                            '5': 'emd'
                        }

                        selected_method = method_map_1d.get(analysis_type, 'unknown')

                        if selected_method == 'unknown':
                            print("    Invalid selection. Please select a valid analysis method (1-5).\n")
                            continue

                        # Generate and display the calling sequence for 1D signal
                        print("\n    Calling sequence:\n")
                        return_values = single_series_parameters[selected_method]['return_values']
                        command = f"    >>> {return_values} = WaLSAtools(signal=INPUT_DATA, time=TIME_ARRAY, method='{selected_method}', **kwargs)"
                        print(command)

                        # Display parameter hints
                        display_parameters_text(selected_method, category='a')
                        return  # Exit the function after successful output

                elif method == '2':  # 3D Datacube
                    while True:
                        print("\n    Analysis Method:")
                        print("    (1) k-omega")
                        print("    (2) POD")
                        print("    (3) Dominant Freq / Mean Power Spectrum")
                        analysis_type = input("    --- Select an analysis method (1-3): ").strip()

                        if analysis_type == '3':  # Sub-method required
                            while True:
                                print("\n    Analysis method for Dominant Freq / Mean Power Spectrum:")
                                print("    (1) FFT")
                                print("    (2) Wavelet")
                                print("    (3) Lomb-Scargle")
                                print("    (4) Welch")
                                sub_method = input("    --- Select a method (1-4): ").strip()

                                sub_method_map = {'1': 'fft', '2': 'wavelet', '3': 'lombscargle', '4': 'welch'}
                                selected_method = sub_method_map.get(sub_method, 'unknown')

                                if selected_method == 'unknown':
                                    print("    Invalid selection. Please select a valid sub-method (1-4).\n")
                                    continue

                                # Generate and display the calling sequence for sub-method
                                print("\n    Calling sequence:\n")
                                return_values = 'dominant_frequency, mean_power, frequency, power_map'
                                command = f"    >>> {return_values} = WaLSAtools(data=INPUT_DATA, time=TIME_ARRAY, averagedpower=True, dominantfreq=True, method='{selected_method}', **kwargs)"
                                print(command)

                                # Display parameter hints
                                display_parameters_text(selected_method, category='a')
                                return  # Exit the function after successful output

                        method_map_3d = {
                            '1': 'k-omega',
                            '2': 'pod'
                        }

                        selected_method = method_map_3d.get(analysis_type, 'unknown')

                        if selected_method == 'unknown':
                            print("    Invalid selection. Please select a valid analysis method (1-3).\n")
                            continue

                        # Generate and display the calling sequence for k-omega/POD
                        print("\n    Calling sequence:\n")
                        return_values = single_series_parameters[selected_method]['return_values']
                        command = f"    >>> {return_values} = WaLSAtools(data1=INPUT_DATA, time=TIME_ARRAY, method='{selected_method}', **kwargs)"
                        print(command)

                        # Display parameter hints
                        display_parameters_text(selected_method, category='a')
                        return  # Exit the function after successful output

        elif category == 'b':  # Cross-correlation
            while True:
                print("\n    Data Type:")
                print("    (1) 1D signal")
                method = input("    --- Select a data type (1): ").strip()

                if method != '1':
                    print("    Invalid selection. Please enter '1'.\n")
                    continue

                while True:
                    print("\n    Analysis Method:")
                    print("    (1) Wavelet")
                    print("    (2) Welch")
                    analysis_type = input("    --- Select an analysis method (1-2): ").strip()

                    cross_correlation_map = {
                        '1': 'wavelet',
                        '2': 'welch'
                    }

                    selected_method = cross_correlation_map.get(analysis_type, 'unknown')

                    if selected_method == 'unknown':
                        print("    Invalid selection. Please select a valid analysis method (1-2).\n")
                        continue

                    # Generate and display the calling sequence for cross-correlation
                    print("\n    Calling sequence:\n")
                    return_values = cross_correlation_parameters[selected_method]['return_values']
                    command = f">>> {return_values} = WaLSAtools(data1=INPUT_DATA1, data2=INPUT_DATA2, time=TIME_ARRAY, method='{selected_method}', **kwargs)"
                    print(command)

                    # Display parameter hints
                    display_parameters_text(selected_method, category='b')
                    return  # Exit the function after successful output


# Jupyter-based interactive function
import ipywidgets as widgets # type: ignore
from IPython.display import display, clear_output, HTML # type: ignore
from .parameter_definitions import display_parameters_html, single_series_parameters, cross_correlation_parameters # type: ignore

# Global flag to prevent multiple observers
is_observer_attached = False

def walsatools_jupyter():
    """Main interactive function for Jupyter Notebook version of WaLSAtools."""
    global is_observer_attached, category, method, analysis_type, sub_method  # Declare global variables for reuse

    # Detach any existing observers and reset the flag
    try:
        detach_observers()
    except NameError:
        pass  # `detach_observers` hasn't been defined yet

    is_observer_attached = False  # Reset observer flag

    # Clear any previous output
    clear_output(wait=True)

    print_logo_and_credits()

    # Recreate widgets to reset state
    category = widgets.Dropdown(
        options=['Select Category', 'Single time series analysis', 'Cross-correlation between two time series'],
        value='Select Category',
        description='Category:'
    )
    method = widgets.Dropdown(
        options=['Select Data Type'],
        value='Select Data Type',
        description='Data Type:'
    )
    analysis_type = widgets.Dropdown(
        options=['Select Method'],
        value='Select Method',
        description='Method:'
    )
    sub_method = widgets.Dropdown(
        options=['Select Sub-method', 'FFT', 'Wavelet', 'Lomb-Scargle', 'Welch'],
        value='Select Sub-method',
        description='Sub-method:',
        layout=widgets.Layout(display='none')  # Initially hidden
    )

    # Persistent output widget
    output = widgets.Output()

    def clear_output_if_unselected(change=None):
        """Clear the output widget if any dropdown menu is unselected."""
        with output:
            if (
                category.value == 'Select Category'
                or method.value == 'Select Data Type'
                or analysis_type.value == 'Select Method'
                or (
                    analysis_type.value == 'Dominant Freq / Mean Power Spectrum'
                    and sub_method.layout.display == 'block'
                    and sub_method.value == 'Select Sub-method'
                )
            ):
                clear_output(wait=True)
                warn = '<div style="font-size: 1.1em; margin-left: 30px; margin-top:15px; margin-bottom: 15px;">Please select appropriate options from all dropdown menus.</div>'
                display(HTML(warn))

    def update_method_options(change=None):
        """Update available Method and Sub-method options."""
        clear_output_if_unselected()  # Ensure the output clears if any dropdown is unselected.
        sub_method.layout.display = 'none'
        sub_method.value = 'Select Sub-method'
        sub_method.options = ['Select Sub-method']

        if category.value == 'Single time series analysis':
            method.options = ['Select Data Type', '1D signal', '3D datacube']
            if method.value == '1D signal':
                analysis_type.options = ['Select Method', 'FFT', 'Wavelet', 'Lomb-Scargle', 'Welch', 'EMD']
            elif method.value == '3D datacube':
                analysis_type.options = ['Select Method', 'k-omega', 'POD', 'Dominant Freq / Mean Power Spectrum']
            else:
                analysis_type.options = ['Select Method']
        elif category.value == 'Cross-correlation between two time series':
            method.options = ['Select Data Type', '1D signal']
            if method.value == '1D signal':
                analysis_type.options = ['Select Method', 'Wavelet', 'Welch']
            else:
                analysis_type.options = ['Select Method']
        else:
            method.options = ['Select Data Type']
            analysis_type.options = ['Select Method']

    def update_sub_method_visibility(change=None):
        """Show or hide the Sub-method dropdown based on conditions."""
        clear_output_if_unselected()  # Ensure the output clears if any dropdown is unselected.
        if (
            category.value == 'Single time series analysis'
            and method.value == '3D datacube'
            and analysis_type.value == 'Dominant Freq / Mean Power Spectrum'
        ):
            sub_method.options = ['Select Sub-method', 'FFT', 'Wavelet', 'Lomb-Scargle', 'Welch']
            sub_method.layout.display = 'block'
        else:
            sub_method.options = ['Select Sub-method']
            sub_method.value = 'Select Sub-method'
            sub_method.layout.display = 'none'

    def update_calling_sequence(change=None):
        """Update the function calling sequence based on user's selection."""
        clear_output_if_unselected()  # Ensure the output clears if any dropdown is unselected.
        if (
            category.value == 'Select Category'
            or method.value == 'Select Data Type'
            or analysis_type.value == 'Select Method'
            or (
                analysis_type.value == 'Dominant Freq / Mean Power Spectrum'
                and sub_method.layout.display == 'block'
                and sub_method.value == 'Select Sub-method'
            )
        ):
            return  # Do nothing until all required fields are properly selected

        with output:
            clear_output(wait=True)

            # Handle Dominant Frequency / Mean Power Spectrum with Sub-method
            if (
                category.value == 'Single time series analysis'
                and method.value == '3D datacube'
                and analysis_type.value == 'Dominant Freq / Mean Power Spectrum'
            ):
                if sub_method.value == 'Select Sub-method':
                    print("Please select a Sub-method.")
                    return

                sub_method_map = {'FFT': 'fft', 'Wavelet': 'wavelet', 'Lomb-Scargle': 'lombscargle', 'Welch': 'welch'}
                selected_method = sub_method_map.get(sub_method.value, 'unknown')
                return_values = 'dominant_frequency, mean_power, frequency, power_map'
                command = f"{return_values} = WaLSAtools(signal=INPUT_DATA, time=TIME_ARRAY, averagedpower=True, dominantfreq=True, method='{selected_method}', **kwargs)"

            # Handle k-omega and POD
            elif (
                category.value == 'Single time series analysis'
                and method.value == '3D datacube'
                and analysis_type.value in ['k-omega', 'POD']
            ):
                method_map = {'k-omega': 'k-omega', 'POD': 'pod'}
                selected_method = method_map.get(analysis_type.value, 'unknown')
                parameter_definitions = single_series_parameters
                return_values = parameter_definitions.get(selected_method, {}).get('return_values', '')
                command = f"{return_values} = WaLSAtools(signal=INPUT_DATA, time=TIME_ARRAY, method='{selected_method}', **kwargs)"

            # Handle Cross-correlation
            elif category.value == 'Cross-correlation between two time series':
                cross_correlation_map = {'Wavelet': 'wavelet', 'Welch': 'welch'}
                selected_method = cross_correlation_map.get(analysis_type.value, 'unknown')
                parameter_definitions = cross_correlation_parameters
                return_values = parameter_definitions.get(selected_method, {}).get('return_values', '')
                command = f"{return_values} = WaLSAtools(data1=INPUT_DATA1, data2=INPUT_DATA2, time=TIME_ARRAY, method='{selected_method}', **kwargs)"

            # Handle Single 1D signal analysis
            elif category.value == 'Single time series analysis' and method.value == '1D signal':
                method_map = {'FFT': 'fft', 'Wavelet': 'wavelet', 'Lomb-Scargle': 'lombscargle', 'Welch': 'welch', 'EMD': 'emd'}
                selected_method = method_map.get(analysis_type.value, 'unknown')
                parameter_definitions = single_series_parameters
                return_values = parameter_definitions.get(selected_method, {}).get('return_values', '')
                command = f"{return_values} = WaLSAtools(signal=INPUT_DATA, time=TIME_ARRAY, method='{selected_method}', **kwargs)"

            else:
                print("Invalid configuration.")
                return

            # Generate and display the command in HTML
            html_code = f"""
            <div style="font-size: 1.2em; margin-left: 30px; margin-top:15px; margin-bottom: 15px;">Calling sequence:</div>
            <div style="display: flex; margin-left: 30px; margin-bottom: 3ch;">
                <span style="color: #222; min-width: 4ch;">>>> </span>
                <pre style="
                    white-space: pre-wrap; 
                    word-wrap: break-word;  
                    color: #01016D; 
                    margin: 0;
                ">{command}</pre>
            </div>
            """
            display(HTML(html_code))
            display_parameters_html(selected_method, category=category.value)

    def attach_observers():
        """Attach observers to the dropdown widgets and ensure no duplicates."""
        global is_observer_attached
        if not is_observer_attached:
            detach_observers()
            category.observe(clear_output_if_unselected, names='value')
            method.observe(clear_output_if_unselected, names='value')
            analysis_type.observe(clear_output_if_unselected, names='value')
            category.observe(update_method_options, names='value')
            method.observe(update_method_options, names='value')
            analysis_type.observe(update_sub_method_visibility, names='value')
            sub_method.observe(update_calling_sequence, names='value')
            analysis_type.observe(update_calling_sequence, names='value')
            is_observer_attached = True

    def detach_observers():
        """Detach all observers to prevent multiple triggers."""
        try:
            category.unobserve(update_method_options, names='value')
            method.unobserve(update_method_options, names='value')
            analysis_type.unobserve(update_sub_method_visibility, names='value')
            sub_method.unobserve(update_calling_sequence, names='value')
            analysis_type.unobserve(update_calling_sequence, names='value')
        except ValueError:
            pass

    attach_observers()
    display(category, method, analysis_type, sub_method, output)


# Unified interactive function for both terminal and Jupyter environments
def interactive():
    if is_notebook():
        walsatools_jupyter()
    else:
        walsatools_terminal()


#!/usr/bin/python
# coding: UTF-8
#
# Author:   Dawid Laszuk
# Contact:  https://github.com/laszukdawid/PyEMD/issues
#
# Edited:   07/07/2017
#
# Feel free to contact for any information.

import logging

import numpy as np

try:
    from scipy.interpolate import SmoothBivariateSpline as SBS
    from scipy.ndimage.filters import maximum_filter
    from scipy.ndimage.morphology import binary_erosion, generate_binary_structure
except ImportError:
    raise ImportError(
        "EMD2D and BEMD are not supported. Feel free to play around and improve them. "
        + "Required depdenecies are in `requriements-extra`."
    )


class EMD2D:
    """
    **Empirical Mode Decomposition** on images.

    **Important** This is an experimental module.
    Experiments performed using this module didn't provide acceptable results,
    either in actual output nor in computation performance. The author is not
    an expert in image processing so it's very likely that the code could
    have been improved. Take your best shot.

    Method decomposes images into 2D representations of loose Intrinsic Mode
    Functions (IMFs).

    The current version of the algorithm detects local extrema, separately
    minima and maxima, and then connects them to create envelopes. These
    are then used to create a mean trend and subtracted from the input.

    Threshold values that control goodness of the decomposition:
        * `mse_thr` --- proto-IMF check whether small mean square error.
        * `mean_thr` --- proto-IMF chekc whether small mean value.
    """

    logger = logging.getLogger(__name__)

    def __init__(self, **config):
        # ProtoIMF related
        self.mse_thr = 0.01
        self.mean_thr = 0.01

        self.FIXE = 0
        self.FIXE_H = 0

        self.MAX_ITERATION = 1000

        # Update based on options
        for key in config.keys():
            if key in self.__dict__.keys():
                self.__dict__[key] = config[key]

    def __call__(self, image, max_imf=-1):
        return self.emd(image, max_imf=max_imf)

    def extract_max_min_spline(self, image):
        """Calculates top and bottom envelopes for image.

        Parameters
        ----------
        image : numpy 2D array

        Returns
        -------
        min_env : numpy 2D array
            Bottom envelope in form of an image.
        max_env : numpy 2D array
            Top envelope in form of an image.
        """

        big_image = self.prepare_image(image)
        big_min_peaks, big_max_peaks = self.find_extrema(big_image)

        # Prepare grid for interpolation. Doesn't seem necessary.
        xi = np.arange(image.shape[0], image.shape[0] * 2)
        yi = np.arange(image.shape[1], image.shape[1] * 2)

        big_min_image_val = big_image[big_min_peaks]
        big_max_image_val = big_image[big_max_peaks]
        min_env = self.spline_points(big_min_peaks[0], big_min_peaks[1], big_min_image_val, xi, yi)
        max_env = self.spline_points(big_max_peaks[0], big_max_peaks[1], big_max_image_val, xi, yi)

        return min_env, max_env

    @classmethod
    def prepare_image(cls, image):
        """Prepares image for edge extrapolation.
        Method bloats image by mirroring it along all axes. This turns
        extrapolation on edges into interpolation within bigger image.

        Parameters
        ----------
        image : numpy 2D array
            Image for which interpolation is required,

        Returns
        -------
        image : numpy 2D array
            Big image based on the input. Grid 3x3 where the center block is input and
            neighbouring panels are respective mirror images.
        """

        # TODO: This is nasty. Instead of bloating whole image and then trying to
        #      find all extrema, it's better to deal directly with indices.
        shape = image.shape
        big_image = np.zeros((shape[0] * 3, shape[1] * 3))

        image_lr = np.fliplr(image)
        image_ud = np.flipud(image)
        image_ud_lr = np.flipud(image_lr)
        image_lr_ud = np.fliplr(image_ud)

        # Fill center with default image
        big_image[shape[0] : 2 * shape[0], shape[1] : 2 * shape[1]] = image

        # Fill left center
        big_image[shape[0] : 2 * shape[0], : shape[1]] = image_lr

        # Fill right center
        big_image[shape[0] : 2 * shape[0], 2 * shape[1] :] = image_lr

        # Fill center top
        big_image[: shape[0], shape[1] : shape[1] * 2] = image_ud

        # Fill center bottom
        big_image[2 * shape[0] :, shape[1] : 2 * shape[1]] = image_ud

        # Fill left top
        big_image[: shape[0], : shape[1]] = image_ud_lr

        # Fill left bottom
        big_image[2 * shape[0] :, : shape[1]] = image_ud_lr

        # Fill right top
        big_image[: shape[0], 2 * shape[1] :] = image_lr_ud

        # Fill right bottom
        big_image[2 * shape[0] :, 2 * shape[1] :] = image_lr_ud

        return big_image

    @classmethod
    def spline_points(cls, X, Y, Z, xi, yi):
        """Interpolates for given set of points"""

        # SBS requires at least m=(kx+1)*(ky+1) points,
        # where kx=ky=3 (default) is the degree of bivariate spline.
        # Thus, if less than 16=(3+1)*(3+1) points, adjust kx & ky.
        spline = SBS(X, Y, Z)

        return spline(xi, yi)

    @classmethod
    def find_extrema(cls, image):
        """
        Finds extrema, both mininma and maxima, based on local maximum filter.
        Returns extrema in form of two rows, where the first and second are
        positions of x and y, respectively.

        Parameters
        ----------
        image : numpy 2D array
            Monochromatic image or any 2D array.

        Returns
        -------
        min_peaks : numpy array
            Minima positions.
        max_peaks : numpy array
            Maxima positions.
        """

        # define an 3x3 neighborhood
        neighborhood = generate_binary_structure(2, 2)

        # apply the local maximum filter; all pixel of maximal value
        # in their neighborhood are set to 1
        local_min = maximum_filter(-image, footprint=neighborhood) == -image
        local_max = maximum_filter(image, footprint=neighborhood) == image

        # can't distinguish between background zero and filter zero
        background = image == 0

        # appear along the bg border (artifact of the local max filter)
        eroded_background = binary_erosion(background, structure=neighborhood, border_value=1)

        # we obtain the final mask, containing only peaks,
        # by removing the background from the local_max mask (xor operation)
        min_peaks = local_min ^ eroded_background
        max_peaks = local_max ^ eroded_background

        min_peaks = local_min
        max_peaks = local_max
        min_peaks[[0, -1], :] = False
        min_peaks[:, [0, -1]] = False
        max_peaks[[0, -1], :] = False
        max_peaks[:, [0, -1]] = False

        min_peaks = np.nonzero(min_peaks)
        max_peaks = np.nonzero(max_peaks)

        return min_peaks, max_peaks

    @classmethod
    def end_condition(cls, image, IMFs):
        """Determins whether decomposition should be stopped.

        Parameters
        ----------
        image : numpy 2D array
            Input image which is decomposed.
        IMFs : numpy 3D array
            Array for which first dimensions relates to respective IMF,
            i.e. (numIMFs, imageX, imageY).
        """
        rec = np.sum(IMFs, axis=0)

        # If reconstruction is perfect, no need for more tests
        if np.allclose(image, rec):
            return True

        return False

    def check_proto_imf(self, proto_imf, proto_imf_prev, mean_env):
        """Check whether passed (proto) IMF is actual IMF.
        Current condition is solely based on checking whether the mean is below threshold.

        Parameters
        ----------
        proto_imf : numpy 2D array
            Current iteration of proto IMF.
        proto_imf_prev : numpy 2D array
            Previous iteration of proto IMF.
        mean_env : numpy 2D array
            Local mean computed from top and bottom envelopes.

        Returns
        -------
        boolean
            Whether current proto IMF is actual IMF.
        """

        # TODO: Sifiting is very sensitive and subtracting const val can often flip
        #      maxima with minima in decompoisition and thus repeating above/below
        #      behaviour. For now, mean_env is checked whether close to zero excluding
        #      its offset.
        if np.all(np.abs(mean_env - mean_env.mean()) < self.mean_thr):
            # if np.all(np.abs(mean_env)<self.mean_thr):
            return True

        # If very little change with sifting
        if np.allclose(proto_imf, proto_imf_prev):
            return True

        # If IMF mean close to zero (below threshold)
        if np.mean(np.abs(proto_imf)) < self.mean_thr:
            return True

        # Everything relatively close to 0
        mse_proto_imf = np.mean(proto_imf * proto_imf)
        if mse_proto_imf < self.mse_thr:
            return True

        return False

    def emd(self, image, max_imf=-1):
        """Performs EMD on input image with specified parameters.

        Parameters
        ----------
        image : numpy 2D array,
            Image which will be decomposed.
        max_imf : int, (default: -1)
            IMF number to which decomposition should be performed.
            Negative value means *all*.

        Returns
        -------
        IMFs : numpy 3D array
            Set of IMFs in form of numpy array where the first dimension
            relates to IMF's ordinary number.
        """
        image_min, image_max = np.min(image), np.max(image)
        offset = image_min
        scale = image_max - image_min

        image_s = (image - offset) / scale

        imf = np.zeros(image.shape)
        imf_old = imf.copy()

        imfNo = 0
        IMF = np.empty((imfNo,) + image.shape)
        notFinished = True

        while notFinished:
            self.logger.debug("IMF -- " + str(imfNo))

            res = image_s - np.sum(IMF[:imfNo], axis=0)
            imf = res.copy()
            mean_env = np.zeros(image.shape)
            stop_sifting = False

            # Counters
            n = 0  # All iterations for current imf.
            n_h = 0  # counts when mean(proto_imf) < threshold

            while not stop_sifting and n < self.MAX_ITERATION:
                n += 1
                self.logger.debug("Iteration: " + str(n))

                min_peaks, max_peaks = self.find_extrema(imf)

                self.logger.debug("min_peaks = %i  |  max_peaks = %i", len(min_peaks[0]), len(max_peaks[0]))
                if len(min_peaks[0]) > 4 and len(max_peaks[0]) > 4:
                    imf_old = imf.copy()
                    imf = imf - mean_env

                    min_env, max_env = self.extract_max_min_spline(imf)

                    mean_env = 0.5 * (min_env + max_env)

                    imf_old = imf.copy()
                    imf = imf - mean_env

                    # Fix number of iterations
                    if self.FIXE:
                        if n >= self.FIXE + 1:
                            stop_sifting = True

                    # Fix number of iterations after number of zero-crossings
                    # and extrema differ at most by one.
                    elif self.FIXE_H:
                        if n == 1:
                            continue
                        if self.check_proto_imf(imf, imf_old, mean_env):
                            n_h += 1
                        else:
                            n_h = 0

                        # STOP if enough n_h
                        if n_h >= self.FIXE_H:
                            stop_sifting = True

                    # Stops after default stopping criteria are met
                    else:
                        if self.check_proto_imf(imf, imf_old, mean_env):
                            stop_sifting = True

                else:
                    notFinished = False
                    stop_sifting = True

            IMF = np.vstack((IMF, imf.copy()[None, :]))
            imfNo += 1

            if self.end_condition(image, IMF) or (max_imf > 0 and imfNo >= max_imf):
                notFinished = False
                break

        res = image_s - np.sum(IMF[:imfNo], axis=0)
        if not np.allclose(res, 0):
            IMF = np.vstack((IMF, res[None, :]))
            imfNo += 1

        IMF = IMF * scale
        IMF[-1] += offset
        return IMF


########################################
if __name__ == "__main__":
    print("Running example on EMD2D")
    PLOT = True

    logging.basicConfig(level=logging.DEBUG)

    # Generate image
    print("Generating image... ", end="")
    rows, cols = 1024, 1024
    row_scale, col_scale = 256, 256
    x = np.arange(rows) / float(row_scale)
    y = np.arange(cols).reshape((-1, 1)) / float(col_scale)

    pi2 = 2 * np.pi
    img = np.zeros((rows, cols))
    img = img + np.sin(2 * pi2 * x) * np.cos(y * 4 * pi2 + 4 * x * pi2)
    img = img + 3 * np.sin(2 * pi2 * x) + 2
    img = img + 5 * x * y + 2 * (y - 0.2) * y
    print("Done")

    # Perform decomposition
    print("Performing decomposition... ", end="")
    emd2d = EMD2D()
    # emd2d.FIXE_H = 5
    IMFs = emd2d.emd(img, max_imf=4)
    imfNo = IMFs.shape[0]
    print("Done")

    if PLOT:
        print("Plotting results... ", end="")
        import pylab as plt

        # Save image for preview
        plt.figure(figsize=(4, 4 * (imfNo + 1)))
        plt.subplot(imfNo + 1, 1, 1)
        plt.imshow(img)
        plt.colorbar()
        plt.title("Input image")

        # Save reconstruction
        for n, imf in enumerate(IMFs):
            plt.subplot(imfNo + 1, 1, n + 2)
            plt.imshow(imf)
            plt.colorbar()
            plt.title("IMF %i" % (n + 1))

        plt.savefig("image_decomp")
        print("Done")
